#include "IhMc.h"
#include "cidx.h"
#include <utility>
#include <iomanip>
#include <algorithm>
#include <assert.h>
#include <cmath>

using std::min;
using std::setprecision;
using std::setw;
using std::ostream;
using std::endl;
using std::cerr;
using std::cout;
using std::string;
using std::vector;


IhMc::IhMc(const Model* ex, int col, double val) : _vec(col) {
  if( ex ) {
    for(unsigned i = 0; i < _vec.size(); ++i) {
      _vec[i] = ex->clone();
    }
  }
}


IhMc::IhMc(const IhMc& wm) : _vec(wm._vec.size()) {
  for(unsigned i = 0; i < _vec.size(); ++i) {
    _vec[i] = wm._vec[i]->clone();
  }
}

IhMc::~IhMc() {
  for(unsigned i = 0; i < _vec.size(); ++i) {
    delete _vec[i];
  }
}

Model*
IhMc::clone() const {
  return new IhMc(*this);
}


Model&
IhMc::operator=(const Model& mm) {
  const IhMc* kmPtr = dynamic_cast<const IhMc*>(&mm);
  assert(kmPtr);
  const IhMc& km = *kmPtr;
  _vec.resize(km._vec.size());
  for(unsigned i = 0; i < _vec.size(); ++i) {
    _vec[i] = km.getElem(i)->clone();
  }
  return *this;
}

void
IhMc::operator+=(const Model& mm) {
  const IhMc* kmPtr = dynamic_cast<const IhMc*>(&mm);
  assert(kmPtr);
  const IhMc& km = *kmPtr;
  assert(_vec.size()==km._vec.size());
  for(unsigned i = 0; i < _vec.size(); ++i) {
    (*_vec[i]) += *(km.getElem(i));
  }
}

void
IhMc::operator*=(const Model& mm) {
  const IhMc* kmPtr = dynamic_cast<const IhMc*>(&mm);
  assert(kmPtr);
  const IhMc& km = *kmPtr;
  assert(_vec.size()==km._vec.size());
  for(unsigned i = 0; i < _vec.size(); ++i) {
    (*_vec[i]) *= *(km.getElem(i));
  }
}

void
IhMc::add(double val) {
  for(unsigned i = 0; i < _vec.size(); ++i) {
    _vec[i]->add(val);
  }
}

void
IhMc::operator*=(double mm) {
  for(unsigned i = 0; i < _vec.size(); ++i) {
    (*_vec[i]) *= mm;
  }
}


void
IhMc::operator-=(const Model& mm) {
  const IhMc* kmPtr = dynamic_cast<const IhMc*>(&mm);
  assert(kmPtr);
  const IhMc& km = *kmPtr;
  assert(_vec.size()==km._vec.size());
  for(unsigned i = 0; i < _vec.size(); ++i) {
    (*_vec[i]) -= *(km.getElem(i));
  }
}

Model*
IhMc::operator-(const Model& mm) const {
  const IhMc* kmPtr = dynamic_cast<const IhMc*>(&mm);
  assert(kmPtr);
  const IhMc& km = *kmPtr;
  assert(_vec.size()==km._vec.size());
  IhMc* res = new IhMc(NULL,_vec.size(),0);
  for(unsigned i = 0; i < _vec.size(); ++i) {
    res->_vec[i] = (*_vec[i]) - *(km.getElem(i));
  }
  return res;
}


void
IhMc::operator/=(int div) {
  for(unsigned i = 0; i < _vec.size(); ++i) {
    (*_vec[i]) /= div;
  }
}

typedef std::pair<unsigned,unsigned> upair_t;

static upair_t
shiftToStart(signed rght, signed left) {
  unsigned wIdx = 0;
  unsigned strIdx = 0;
  if(left < rght) {
    assert((-left+rght) >= 0);
    wIdx = -left+rght;
    strIdx = rght;
  } else {
    strIdx = left;
    wIdx = 0;
  }
  return upair_t(wIdx,strIdx);
}


void 
IhMc::update(const string& str, int cStr_coord,int offset) {
  assert(offset >=0 && offset < (signed)getWidth());
  int lpos = 0;
  unsigned rght = getWidth();
  upair_t pr = shiftToStart(lpos,cStr_coord);
  unsigned rpos = str.size();
  for(unsigned wIdx = pr.first + offset, strIdx = pr.second; wIdx < rght && strIdx < rpos; 
      ++wIdx,++strIdx) {
    assert(strIdx >= 0);
    //cout<<wIdx<<" "<<strIdx<<" "<<str[strIdx]<<endl;
    _vec[wIdx]->update(str,strIdx,strIdx+1);

    // if offset is specified means we're updating specific element
    // otherwise if -1, means we're updating a complete interval
    if( offset != -1 ){ 
      break;
    }
  }
}

double
IhMc::getLogProb(const string& str, int start, int rpos) const {
  const int pos = (signed)getWidth()-(rpos-start)-1;
  assert(pos>=0 && pos < (signed)getWidth());
  double sum = 0;
  for(signed i = pos; i < (signed)getWidth() && ((start+i)<=rpos); ++i) {
    double val = getVal(str,start+i,i);
    assert(val>=0&&val<=1);
    val = val==0 ? 0.000001:val;
    sum += log2(val);
  }
  return sum;
}


double 
IhMc::getVal(const string& str, int strIdx, unsigned wIdx) const {
  assert(strIdx >= 0 && strIdx <= (signed)str.size());
  assert(wIdx >= 0 && wIdx < getWidth());
  const double retval = _vec[wIdx]->getVal(str,strIdx);
  return retval;
}


void
IhMc::normalize() {
  const unsigned num_pos = getWidth();
  for(unsigned pos = 0; pos < num_pos; ++pos) {
    _vec[pos]->normalize();
  }
}

void
IhMc::gnuPlot(std::ostream& os) const {
  for(unsigned i = 0; i < _vec.size(); ++i) {
    Model* mod = _vec[i];
    mod->gnuPlot(os);
  }
}



ostream&
operator<<(ostream& os, const IhMc& mat) {
  for(unsigned pos = 0; pos < mat.getWidth(); ++pos) {
    os<<"pos: "<<pos<<endl;
    mat.getElem(pos)->print(os);
    os<<endl;
  }
  return os;
}

void
IhMc::write(ostream& os) const {
  for(unsigned pos = 0; pos < getWidth(); ++pos) {
    getElem(pos)->write(os);
  }
}

void
IhMc::read(std::istream& is) {
  for(unsigned pos = 0; pos < getWidth(); ++pos) {
    Model* mc = getElem(pos);
    mc->read(is);
  }
}



ostream&
IhMc::print(ostream& os) const {
  os<<*this;
  return os;
}
