#include "MarkovChain.h"
#include "cidx.h"
#include <list>
#include <stdlib.h>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <cmath>
#include <iostream>
#include <cctype>

extern bool GLOBAL_DEBUG1;

using namespace std;

static int 
keyVal(const string&); 
static int
bitVal(char ic); 

MarkovChain::MarkovChain(int chain_len, double val, int output_size) : _counter(output_size), _cOrder(chain_len) {
  // When 0 order markov chain, chain_len = 0, but it's still
  // stored in a 2-d array, just one of the columns will be used

  //chain_len = chain_len <= 0 ? 1 : chain_len;
  const int prevSize = (int)pow((float)output_size,(int)chain_len);
  for(unsigned i = 0; i < (unsigned)output_size; ++i) {
    _counter[i].resize(prevSize,val);
  }
}


MarkovChain::MarkovChain(const vector<vector<double> >& counter, int prev_code) : _counter(counter.size()), _cOrder(0) {
  const int chain_len = 0;
  const float output_size = (float)counter.size();
  const int prevSize = (int)pow(output_size,(int)chain_len);
  for(unsigned output = 0; output < counter.size(); ++output) {
    _counter[output].resize(prevSize);
    _counter[output][0] = counter[output][prev_code];
  }
}

MarkovChain::MarkovChain(const MarkovChain& mc) : _counter(mc._counter.size()), _cOrder(mc._cOrder) {
  for(unsigned output = 0; output < _counter.size(); ++output) {
    const unsigned clen = mc._counter[output].size();
    _counter[output].resize(clen,0);
    for(unsigned j = 0; j < clen; ++j) {
      _counter[output][j] = mc._counter[output][j];
    }
  }
}

Model*
MarkovChain::clone() const {
  return new MarkovChain(*this);
}


Model&
MarkovChain::operator=(const Model& mm) {
  const MarkovChain* mcPtr = dynamic_cast<const MarkovChain*>(&mm);
  assert(mcPtr);
  const MarkovChain& mc = *mcPtr;
  _counter.resize(mc._counter.size());
  _cOrder = mc._cOrder; 
  for(unsigned output = 0; output < _counter.size(); ++output) {
    const unsigned clen = mc._counter[output].size();
    _counter[output].resize(clen,0);
    for(unsigned j = 0; j < clen; ++j) {
      _counter[output][j] = mc._counter[output][j];
    }
  }
  return *this;
}

Model* 
MarkovChain::operator-(const Model& mm) const {
  const MarkovChain* mcPtr = dynamic_cast<const MarkovChain*>(&mm);
  assert(mcPtr);
  const MarkovChain& mc = *mcPtr;
  MarkovChain* res = new MarkovChain(*this);
  for(unsigned output = 0; output < _counter.size(); ++output) {
    const unsigned clen = mc._counter[output].size();
    for(unsigned j = 0; j < clen; ++j) {
      res->_counter[output][j] = fabs(_counter[output][j] - mc._counter[output][j]);
      assert(res->_counter[output][j]>=0);
    }
  }
  return res;
}

void
MarkovChain::operator-=(const Model& mm) {
  const MarkovChain* mcPtr = dynamic_cast<const MarkovChain*>(&mm);
  assert(mcPtr);
  const MarkovChain& mc = *mcPtr;
  for(unsigned output = 0; output < _counter.size(); ++output) {
    const unsigned clen = mc._counter[output].size();
    for(unsigned j = 0; j < clen; ++j) {
      _counter[output][j] -= mc._counter[output][j];
      //assert(_counter[output][j]>=0);
      if(_counter[output][j]<0) {
	_counter[output][j]= 0;
      }
    }
  }
}

void
MarkovChain::operator*=(const Model& mm) {
  const MarkovChain* mcPtr = dynamic_cast<const MarkovChain*>(&mm);
  assert(mcPtr);
  const MarkovChain& mc = *mcPtr;
  for(unsigned output = 0; output < _counter.size(); ++output) {
    const unsigned clen = mc._counter[output].size();
    for(unsigned j = 0; j < clen; ++j) {
      _counter[output][j] *= mc._counter[output][j];
      assert(_counter[output][j]>=0);
    }
  }
}

void
MarkovChain::operator*=(double mm) {
  for(unsigned output = 0; output < _counter.size(); ++output) {
    const unsigned clen = _counter[output].size();
    for(unsigned j = 0; j < clen; ++j) {
      _counter[output][j] *= mm;
      assert(_counter[output][j]>=0);
    }
  }
}


void
MarkovChain::add(double val) {
  assert(val > 0);
  for(unsigned output = 0; output < _counter.size(); ++output) {
    const unsigned clen = _counter[output].size();
    for(unsigned j = 0; j < clen; ++j) {
      _counter[output][j] += val;
    }
  }
}
void
MarkovChain::operator+=(const Model& mm) {
  const MarkovChain* mcPtr = dynamic_cast<const MarkovChain*>(&mm);
  assert(mcPtr);
  const MarkovChain& mc = *mcPtr;
  assert(_cOrder == mc._cOrder); 
  for(unsigned output = 0; output < _counter.size(); ++output) {
    const unsigned clen = mc._counter[output].size();
    for(unsigned j = 0; j < clen; ++j) {
      _counter[output][j] += mc._counter[output][j];
    }
  }
}

void
MarkovChain::operator/=(int div) {
  for(unsigned output = 0; output < _counter.size(); ++output) {
    const unsigned clen = _counter[output].size();
    for(unsigned j = 0; j < clen; ++j) {
      _counter[output][j] /= div;
    }
  }
}


int 
MarkovChain::getTail(const string& str, int pos) const {
  int prev = 0;
  if( _cOrder > 0 ) {
    string hxmer(_cOrder,'\0');
    for(signed k = _cOrder-1, sidx = 1; k >=0; ++sidx ) {
      const int idx = sidx;
      if( ((pos - idx) - 1) >= 0 ) {
	assert((pos-idx)-1 < (signed)str.size());
	const char tchr = str[(pos-idx)-1];
	if( tchr != '-' ) {
	  assert(k < (signed)hxmer.size());
	  hxmer[k] = tchr;
	  --k;
	}
      } else {
	prev = -1;
	break;
      }
    }
    prev = keyVal(hxmer);
  }
  return prev;
}


double
MarkovChain::getVal(const string& str, int pos, unsigned) const {
  double retval = 0;
  assert(pos-1 < (signed)str.size() && (pos-1)>=0);
  int prev = getTail(str,pos);
  if(prev < 0) {
    cerr<<"non nucleotide character MarkovChainOrderN: need to handle this..."<<pos<<" "<<prev<<endl;
    prev = keyVal("a");
  }
  int next = bioinfo::c2idx(str[pos-1],'a');
  retval = getCount(prev,next);
  return retval;
}

void
MarkovChain::update(const string& str, int left, int rght) {
  int start = left;
  for(signed iter = start; iter < rght && iter <= (signed)str.size(); ++iter) {
    int prev = getTail(str,iter);
    const char x = str[iter-1];
    if( std::isupper(x) ) {
      const int idx = bioinfo::c2idx(x);
      if( prev != -1 && idx != -1 ) {
	count(prev,idx);
      }
    }
  }
}

void
MarkovChain::count(int pIdx, int oIdx) {
  assert(oIdx>=0 && oIdx<(signed)_counter.size());
  // hxmer includes ambiguous base codes
  // which are assumed to be exluded from the alphabet
  if( pIdx != -1 ) {
    assert( pIdx < (signed)_counter[oIdx].size());
    _counter[oIdx][pIdx]++;
  }
}

double
MarkovChain::getCount(int prev, int next) const {
  assert(next < (signed)_counter.size());
  assert( prev < (signed)_counter[0].size());
  return _counter[next][prev];
}

void
MarkovChain::normalize() {
  unsigned output_size = _counter.size();
  assert(output_size>0);
  for(unsigned kmer = 0; kmer < _counter[0].size(); ++kmer) {
    double tot = 0;
    for(unsigned base = 0; base < output_size; ++base) {
      tot += _counter[base][kmer];
    }
    if(tot>0) {
      //double dsum = 0;
      for(unsigned base = 0; base < output_size; ++base) {
	_counter[base][kmer] /= tot;
	assert( _counter[base][kmer] <= 1);
	assert( _counter[base][kmer] >= 0);
	//dsum += _counter[base][kmer];
      }
    } else {
      //   cerr<<"hmmm no data found: "<<kmer<<endl;
    }
  }
}

std::ostream&
MarkovChain::print(std::ostream& os) const {
  unsigned output_size = _counter.size();
  assert(output_size>0);
  for(unsigned base = 0; base < output_size; ++base) {
    for(unsigned kmer = 0; kmer < _counter[base].size(); ++kmer) {
      os<<setw(5)<<setprecision(2);
      os<<kmer<<":"<<_counter[base][kmer]<<" ";
      if( kmer > 0 && (kmer % 20)==0 ) {
	os<<endl;
      }
    }
    os<<endl<<endl;
  }
  return os;
}

void
MarkovChain::gnuPlot(std::ostream& os) const {
  unsigned output_size = _counter.size();
  assert(output_size>0);
  for(unsigned kmer = 0, kid = 0; kmer < _counter[0].size(); ++kmer, kid += output_size) {
    for(unsigned base = 0; base < output_size; ++base) {
      os<<kid+base<<" "<<_counter[base][kmer]<<endl;
    }
  }
  //os<<endl<<endl;
}

void
MarkovChain::write(std::ostream& os) const {
  unsigned output_size = _counter.size();
  assert(output_size>0);
  for(unsigned base = 0; base < output_size; ++base) {
    for(unsigned kmer = 0; kmer < _counter[base].size(); ++kmer) {
      os<<_counter[base][kmer];
      if( kmer+1 < _counter[base].size() )
	os<<" ";
    }
    os<<endl;
  }
}

void
MarkovChain::read(std::istream& is) {
  unsigned output_size = _counter.size();
  assert(output_size>0);
  for(unsigned base = 0; base < output_size; ++base) {
    for(unsigned kmer = 0; kmer < _counter[base].size(); ++kmer) {
      is>>_counter[base][kmer];
    }
  }
}


int
MarkovChain::getOrder() const {
  return _cOrder;
}

int
bitVal(char ic) {
  char c = tolower(ic);
  switch(c) {
  case 'a':
    return 0x0;
  case 'c':
    return 0x1;
  case 'g':
    return 0x2;
  case 't':
    return 0x3;
  default:
    //cerr<<"non a,c,g,t code -skipping"<<endl;
    return 4;
  }
}


#if 0
static double
log2(double val) {
		return log(val)/log(2.0);
}
#endif

double
MarkovChain::getLogProb(const string& str, int left, int rght) const {
  double sum = 0;
  for(int i = left; i <= rght; ++i) {
    double val = getVal(str,i);
    assert(val >= 0 && val <= 1);
    val = val==0?0.000001:val;
    sum += log2(val);
  }
  return sum;
}


int 
keyVal(const string& str) {
  int result = 0;
  for(unsigned i = 0; i < str.length(); ++i) {
    const int val = bitVal(str[i]);
    if( val >= 4 || val < 0) {
      result = -1;
      break;
    }
    const float ex = pow((float)4,(int)i);
    result += (val*(int)ex);
  }
  return result;
}

MarkovChain*
MarkovChain::makeOrder0(const string& str, int pos) const {
  assert(pos-1 < (signed)str.size());
  int prev = getTail(str,pos);
  if( prev != -1) {
    return new MarkovChain(_counter,prev);
  } 
  assert(pos>=0);
  return NULL;
}
