#include "BaumWelch.h"
#include "RateHmm.h"
#include <list>
#include <cmath>

using std::string;
using std::vector;
using std::list;
using std::cout;
using std::endl;

#if _RUNTIME_DEBUG >= 1
extern int DEBUG;
#endif

BaumWelch::BaumWelch(RateHmm& hmm, PhyloModel& pm, MultiModel& mm) : _hmm(hmm), _phyloModel(pm), _nucDistr(mm), _prob(0) {
  const unsigned num_states = _hmm.getNumStates();
  _alpha.resize(num_states);
  _beta.resize(num_states);
  //for(int i = 0; i < NumStates(); ++i) {
    //_alpha[i].resize(trainLen+1);
    //_beta[i].resize(trainLen+1);
  //}
}

BaumWelch::~BaumWelch() {
}

void
BaumWelch::train(const vector<string*>& data, unsigned beg, unsigned end, int init_phase, int end_phase) {
  vector<double> Qi;
  const int trainLen = (end-beg)+1;
  for(unsigned i = 0; i < _hmm.getNumStates(); ++i) {
    _alpha[i].resize(trainLen+1,-1);
    _beta[i].resize(trainLen+1,-1);
  }
  bool isCycle = !(end_phase > 2); // implies donor or acceptor, with >2 position specific positions
  _prob = Forward(data,beg,end,init_phase,Qi,isCycle);
  if( _prob != 0 ) {
    Backward(data,beg,end,end_phase,Qi,isCycle);
    Estimate(data,beg,end,init_phase,isCycle);
  } else {
    cout<<"Failded to train probably hit an inframe stop codon "<<beg<<" "<<end<<" "<<init_phase<<endl;
  }
}

#if 0
double
BaumWelch::getModelProb(const vector<string*>& data, unsigned beg, unsigned end, int init_phase) {
  vector<double> Qi;
  const double prob = Forward(data,beg,end,init_phase);
  return prob;
}
#endif

//inline static double log2(double val) {
   //return (log10(val)/log10(2));
//}

double BaumWelch::getLastProb() const { return _prob; }

void
BaumWelch::Estimate(const vector<string*>& data, unsigned beg, unsigned end, int init_phase, bool isCycle) {
  const unsigned num_states = _hmm.getNumStates();
  const unsigned csize = data.size();
  int phase = init_phase == -1 ? 0 : init_phase;
  //const unsigned trainLen = (end-beg)+1;
  for(unsigned i = 0; i < num_states; ++i) {
    string column(csize,'-');
    fillColumn(column,data,beg);
    RateState& state = _hmm.getState(i);
    const double rate = state.getRate();
    const double so_prob = _phyloModel.getProb(data,beg,phase,column,&_nucDistr,rate);
    state.update1(so_prob*_beta[i][1]);
  }
  for(unsigned j = beg+1; j <= end; ++j) {
    //advance past initial state
    if( init_phase != -1 ) {
      if(isCycle) {
	phase = (phase+1>2) ? 0 : phase+1;
      } else {
	++phase;
      }
    }
    const int posIdx = j-beg;
    string column(csize,'-');
    fillColumn(column,data,j);
    for(unsigned i = 0; i < num_states; ++i) {
      RateState& state = _hmm.getState(i);
      const double rate = state.getRate();
      list<RateTrans*>::iterator iter = state.getOutLst().begin();
      const list<RateTrans*>::iterator stop = state.getOutLst().end();
      for(; iter != stop; ++iter) {
	RateTrans* trans = *iter;
	int la = trans->getLeftId();
	int rb = trans->getRghtId();
	const double so_prob = _phyloModel.getProb(data,j,phase,column,&_nucDistr,rate);
	trans->update1(_alpha[la][posIdx]*trans->getProb()*so_prob*_beta[rb][posIdx+1]);
      }
    }
  }
}

void
BaumWelch::Backward(const vector<string*>& data, unsigned beg, unsigned end, int init_phase, const vector<double>& Qi, bool isCycle) {
  const unsigned csize = data.size();
  int phase = init_phase == -1 ? 0 : init_phase;
  const unsigned num_states = _hmm.getNumStates();
  vector<double> beta_prime(num_states);
  const unsigned trainLen = (end-beg)+1;
  for(unsigned i = 0; i < num_states; ++i)  {
    _beta[i][trainLen-1] = 1.0;
    beta_prime[i] = -1.0;
  }
  for(signed j = (int)end-1; j >= (int)beg; --j) {
    const int posIdx = j-beg;
    string column(csize,'-');
    fillColumn(column,data,j);
    for(int i = (int)num_states-1; i >= 0; --i) {
      RateState& state = _hmm.getState(i);
      const double rate = state.getRate();
      double prob_sum = 0;
      assert(!state.getOutLst().empty());
      list<RateTrans*>::iterator iter = state.getOutLst().begin();
      const list<RateTrans*>::iterator stop = state.getOutLst().end();
      for(; iter != stop; ++iter) {
	RateTrans* trans = *iter;
	const int ti = trans->getRghtId();
	const double so_prob = _phyloModel.getProb(data,j+1,phase,column,&_nucDistr,rate);
	assert(_beta[ti][posIdx+1] != -1.0);
	prob_sum += (_beta[ti][posIdx+1]*trans->getProb()*so_prob);
      }
      beta_prime[i] = prob_sum;
    }
    assert(Qi[posIdx] > 0);
    for(unsigned i = 0; i < num_states; ++i) {
      _beta[i][posIdx] = beta_prime[i]/Qi[posIdx];
      //cout<<"beta: "<<i<<" pi "<<posIdx<<" "<<_beta[i][posIdx]<<" "<<beta_prime[i]<<" "<<Qi[posIdx]<<endl;
    }
    if( init_phase != -1 ) {
      if( isCycle ) {
	phase = (phase-1<0) ? 2 : phase-1;
      } else {
	--phase;
      }
    }
  }
}

double
BaumWelch::Forward(const vector<string*>& data, unsigned beg, unsigned end, int init_phase, vector<double>& Qi, bool isCycle) {
  const unsigned csize = data.size();
  int phase = init_phase == -1 ? 0 : init_phase;
  const int trainLen = (end-beg)+1;
  Qi.resize(trainLen+1,0);
  //double prob_sum = 0;
  const unsigned num_states = _hmm.getNumStates();
  vector<double> alpha_prime(num_states);
  for(unsigned i = 0; i < num_states; ++i) {
    const RateState& state = _hmm.getState(i);
    const double rate = state.getRate();
    const double iprob = state.getInitProb();
    string column(csize,'-');
    fillColumn(column,data,beg);
    const double so_prob = _phyloModel.getProb(data,beg,phase,column,&_nucDistr,rate);
    alpha_prime[i] += (iprob*so_prob);
    Qi[0] += alpha_prime[i];
  }
  for(unsigned i = 0; i < num_states; ++i) {
    _alpha[i][0] = alpha_prime[i]/Qi[0];
    //cout<<"alpha: "<<i<<" pi "<<0<<" "<<_alpha[i][0]<<" "<<alpha_prime[i]<<" "<<Qi[0]<<endl;
  }
  double prob = log2(Qi[0])/trainLen; //((-1.0/trainLen)+log2(Qi[0]));

  for(unsigned pos = beg+1; pos <= end; ++pos) {
    //advance phase from initial state
    if( init_phase != -1 ) {
      if( isCycle ) {
	phase = (phase+1>2) ? 0 : phase+1;
      } else {
	++phase;
      }
    }
    const int posIdx = pos-beg;
    string column(csize,'-');
    fillColumn(column,data,pos);
    vector<double> alpha_prime(num_states,0);
    for(unsigned i = 0; i < num_states; ++i) {
      RateState& state = _hmm.getState(i);
      double prob_sum = 0;
      const double rate = state.getRate();
      list<RateTrans*>::iterator iter = state.getInLst().begin();
      assert(!state.getInLst().empty());
      const list<RateTrans*>::iterator stop = state.getInLst().end();
      //if( ((i+1)%20)==0 ) {
	//assert(state.getInLst().size()==20);
      //} else {
	//assert(state.getInLst().size()==4);
      //}
      for(; iter != stop; ++iter) {
	RateTrans* trans = *iter;
	const int ti = trans->getLeftId();
	const double so_prob = _phyloModel.getProb(data,pos,phase,column,&_nucDistr,rate);
	prob_sum += (_alpha[ti][posIdx-1]*trans->getProb()*so_prob);
      }
      alpha_prime[i] = prob_sum;
      Qi[posIdx] += prob_sum;
    }
    if( Qi[posIdx] <= 0 ) {
      prob = 0;
      break;
    }
    assert(Qi[posIdx] > 0.0);
    for(unsigned i = 0; i < num_states; ++i) {
      _alpha[i][posIdx] = alpha_prime[i]/Qi[posIdx];
      //cout<<"alpha: "<<i<<" pi "<<posIdx<<" "<<_alpha[i][posIdx]<<" "<<alpha_prime[i]<<" "<<Qi[posIdx]<<endl;
    }
    prob += log2(Qi[posIdx])/trainLen; //((-1.0/trainLen)+log2(Qi[posIdx]));
  }
  return prob;
}
