#include "InitNucDistr.h"
#include "MarkovChain.h"
#include "IhMc.h"
#include "ParseHeader.h"
#include <vector>
#include <list>
#include <fstream>
#include <stdlib.h>
#include <math.h>
#include <stdexcept>

using std::pair;
using std::cerr;
using std::endl;
using std::cout;
using std::string;
using std::vector;
using std::list;
using std::ofstream;
using std::ifstream;
using std::runtime_error;

extern bool GDEBUG;
static bool DEBUG=false;

InitNucDistr::InitNucDistr() : _dnCnt(0,0) {
  for(unsigned i = 0; i < (unsigned)ModelType::eCnameEnd; ++i) {
    cname_t mn = (cname_t)i;
    createModel(mn);
  }
}

InitNucDistr::~InitNucDistr() {
  del();
}

void
InitNucDistr::del() {
  MultiModelHashT::iterator iter = _mm.begin();
  const MultiModelHashT::iterator end = _mm.end();
  for(; iter != end; ++iter) {
    MultiModel* md = (*iter).second;
    delete md;
  }
  PhyloModelHashT::iterator piter = _pmSlow.begin();
  PhyloModelHashT::iterator pend = _pmSlow.end();
  for(; piter != pend; ++piter) {
    PhyloModel* md = (*piter).second;
    delete md;
  }
  piter = _pmFast.begin();
  pend = _pmFast.end();
  for(; piter != pend; ++piter) {
    PhyloModel* md = (*piter).second;
    delete md;
  }
}

#if 0
void 
InitNucDistr::normalize() {
  MultiModelHashT::iterator iter = _mm.begin();
  const MultiModelHashT::iterator end = _mm.end();
  for(; iter != end; ++iter) {
    cname_t id = (*iter).first;
    assert(id != ModelType::eCnameEnd);
    MultiModel* md = (*iter).second;
    md->normalize();
  }
}
#endif

spair_t
InitNucDistr::compScore(const vector<string*>& svec, int pos,  int time, cname_t typ) {
  const unsigned num_org = svec.size();
  assert( pos < (signed)svec[0]->length()); 
  string column(num_org,'-');
  fillColumn(column,svec,pos);
  //cout<<column<<endl;
  bool status = true;
  double res = 0;
  MultiModel* md = _mm[typ];
  PhyloModel* pmf = _pmFast[typ];
  assert(pmf);
  double probF = pmf->getProb(svec,pos,time,column,md);

  PhyloModel* pms = _pmSlow[typ];
  assert(pms);
  double probS = pms->getProb(svec,pos,time,column,md);

  if(probF == -1 ) {
    return spair_t(status,0);
  }
  if( probF==0 ) {
    GDEBUG=false;
    probF=0.0001;
    status = false;
    
  }
  if( probS==0 ) {
    //cout<<"probS: "<<column<<endl;
    probS=0.0001;
    status = false;
  }
  assert(probS>0&&probF>0);
  res = log2(probS)-log2(probF);
  status = false;
  return spair_t(status,res);
}

InitNucDistr::cname_t
InitNucDistr::getModel(const SeqLabel& leftLabel, const SeqLabel& rghtLabel, const ExonDescrip& exonDescrip, unsigned pos) {
  assert((signed)pos > leftLabel.getRght());
  cname_t mn = ModelType::eCnameEnd;
  if( rghtLabel.isIn(pos) ) {
    mn = ModelType::sl2cnamePos(rghtLabel);
  } else {
    // exon
    bool isExon = false;
    if( (SeqLabel::isAcc(leftLabel.getLabel()) || leftLabel.getLabel() == SeqLabel::eBeg) && 
	(SeqLabel::isDon(rghtLabel.getLabel()) || rghtLabel.getLabel() == SeqLabel::eEnd) ) {
      isExon = true;
    // retained intron
    } else if( SeqLabel::isDon(leftLabel.getLabel()) &&
	     SeqLabel::isAcc(rghtLabel.getLabel())) {
      isExon = true;
    }
    // multiple donor region
    else if( SeqLabel::isDon(leftLabel.getLabel()) &&
	     SeqLabel::isDon(rghtLabel.getLabel()) ) {
      isExon = true;
    }
    // multiple acceptor region
    else if( SeqLabel::isAcc(leftLabel.getLabel()) &&
	     SeqLabel::isAcc(rghtLabel.getLabel()) ) {
      isExon = true;
    } else {
      mn = ModelType::sl2cnameInv(leftLabel,rghtLabel,false);
    }
    if(isExon) {
      const int rpos = exonDescrip.getTargPos(pos);
      const bool isPrc = exonDescrip.isProteinCoding(rpos,leftLabel,rghtLabel);
      mn = ModelType::sl2cnameInv(leftLabel,rghtLabel,isPrc);
    }
  }
  return mn;
}

void
InitNucDistr::updateModel(const vector<string*>& svec, const SeqLabel& leftLabel, const SeqLabel& rghtLabel,
			  const ExonDescrip& targetCoords) {
  const string& str = *svec[0];
  for(signed pos = leftLabel.getRght()+1; pos <= rghtLabel.getRght(); ++pos) {
    cname_t mn = getModel(leftLabel,rghtLabel,targetCoords,pos);
    int relPos = -1;
    int frame=-1;
    if( mn != ModelType::eCnameEnd ) {
      if(DEBUG) cout<<"slabel "<<pos<<": "<<mn<<endl;
      if( rghtLabel.isIn(pos) ) {
	if( (signed)SeqLabel::getSigWidth(rghtLabel.getLabel()) != rghtLabel.getLength()) {
	  cout<<"skip, this splice site includes insertions: "<<rghtLabel.getLength()<<" "<<rghtLabel<<endl;
	  continue;
	} else {
	  //cout<<"don't skip here: "<<rghtLabel<<endl;
	}
	// get position in signal
	relPos = pos-rghtLabel.getLeft() ;
	const int left = rghtLabel.getLeft();
	bool fail = false;
	char c1='x',c2='x';
	if( relPos == 20 && SeqLabel::isAcc(rghtLabel.getLabel())) {
	  c1 = tolower(str[left+relPos-1]);
	  c2 = tolower(str[left+relPos]);
	  if(c1!='a') {
	    fail=true;
	  }
	}
	if( relPos == 21 && SeqLabel::isAcc(rghtLabel.getLabel())) {
	  c1 = tolower(str[left+relPos-2]);
	  c2 = tolower(str[left+relPos-1]);
	  if( c2 !='g') {
	    fail=true;
	  }
	}
	if( relPos == 3 && SeqLabel::isDon(rghtLabel.getLabel())) {
	  c1 = tolower(str[left+relPos-1]);
	  c2 = tolower(str[left+relPos]);
	  if( c1 != 'g') 
	    fail=true;
	}
	if( relPos == 4 && SeqLabel::isDon(rghtLabel.getLabel())) {
	  c1 = tolower(str[left+relPos-2]);
	  c2 = tolower(str[left+relPos-1]);
	  if( c2 != 't' && c2 != 'c') 
	    fail=true;
	  if( c1=='g'&&c2=='c') {
	    _dnCnt.first++;
	  }
	  _dnCnt.second++;
	}
	if( fail ) {
	  string sub = str.substr(left-1,(rghtLabel.getRght()-left)+1);
	  //cout<<"funky splice site hmm: ["<<c1<<"] ["<<c2<<"] "<<sub<<endl;
	  //cout<<"coords: "<<left<<" "<<relPos<<" "<<rghtLabel<<endl;
	}
      }
      else if(  ModelType::isPrcExon(mn) ) {
	const int rpos = targetCoords.getTargPos(pos);
	assert(rpos != -1);
	frame=targetCoords.getFrame(rpos,leftLabel,rghtLabel);
	relPos=frame;
	if( pos-3 >= 0) {
	  string codon = str.substr(pos-3,3);
	  makelower(codon);
	  if(frame == 2) {
	    if( codon=="tga" || codon=="taa"||codon=="tag") {
	      ExonDescrip::ipair_t ip = targetCoords.getStartStop();	
	      cout<<"start/stop: "<<ip.first<<" "<<ip.second<<endl;
	      cout<<"stop codon: "<<codon<<" "<<leftLabel<<" "<<rghtLabel<<" pos: "<<pos<<endl;
	      throw std::runtime_error("issues with the stop codon skip");
	      //const int tmp=targetCoords.getFrame(rpos,leftLabel,rghtLabel);
	      //cout<<"stop codon quit"<<endl;
	    }
	  }
	}
      } else {
	relPos = pos+1;
      }
      assert(relPos != -1);
      if( mn != ModelType::eCnameEnd ) {
	MultiModel* mod = _mm[mn];
	mod->update(svec,pos,relPos);
      }
    }
  }
}

struct TCmp {
  bool operator()(const SeqLabel* a, const SeqLabel* b) const {
    return a->getLeft()<b->getLeft();
  }
};

void
InitNucDistr::createModel(cname_t mn) {
  MultiModel* md = NULL;
  const MarkovChain* mc1 = new MarkovChain(1);
  const MarkovChain* mc2 = new MarkovChain(2);
  const ParamFile* pf = ParamFile::getInstance();
  const OrgIdx& oidx = pf->getOrgIdx();
  unsigned osize = oidx.size();
  switch(mn) {
  case ModelType::eAdon: md = new MultiModel(new IhMc(mc1,9,0.0),osize); break;
  case ModelType::eAacc: md = new MultiModel(new IhMc(mc1,24,0.0),osize); break;
  case ModelType::eCdon: md = new MultiModel(new IhMc(mc1,9,0.0),osize); break;
  case ModelType::eCacc: md = new MultiModel(new IhMc(mc1,24,0.0),osize); break;

  case ModelType::eAintronDn: md = new MultiModel(new MarkovChain(2),osize); break;
  case ModelType::eCintronDn: md = new MultiModel(new MarkovChain(2),osize); break;
  case ModelType::eAintronUp: md = new MultiModel(new MarkovChain(2),osize); break;
  case ModelType::eCintronUp: md = new MultiModel(new MarkovChain(2),osize); break;

    // 3-periodic non-homogenious 2nd order MC
  case ModelType::eAexonPrc: md = new MultiModel(new IhMc(mc2,3,0.0),osize); break;
  case ModelType::eCexonPrc: md = new MultiModel(new IhMc(mc2,3,0.0),osize); break;
  case ModelType::eAexonNpc: md = new MultiModel(new MarkovChain(2),osize); break;
  case ModelType::eCexonNpc: md = new MultiModel(new MarkovChain(2),osize); break;
  default: 
    cerr<<"InitNucDistr: no model used for: "<<ModelType::cname2str(mn)<<" "<<mn<<endl;
    break;
  } 
  delete mc1;
  delete mc2;
  if( md ) {
    _mm[mn] = md;
  }
}

void
InitNucDistr::createPhyloModel(cname_t mn, double nmrate) {
  int width = -1; 
  double smP = -1;
  bool useCache = true;
  const double fmrate = 1.0;
  switch(mn) {
  case ModelType::eAdon: width = 9; smP = fmrate; break;
  case ModelType::eCdon: width = 9; smP = fmrate; break;
  case ModelType::eAacc: width = 24; smP = fmrate;break;
  case ModelType::eCacc: width = 24; smP = fmrate;break;

  case ModelType::eAintronDn: width = 1; smP = nmrate; useCache = true; break;
  case ModelType::eCintronDn: width = 1; smP = nmrate; useCache = true; break;
  case ModelType::eAintronUp: width = 1; smP = nmrate; useCache = true; break;
  case ModelType::eCintronUp: width = 1; smP = nmrate; useCache = true; break;

    // 3-periodic non-homogenious 2nd order MC
  case ModelType::eAexonPrc: width = 3; smP = nmrate; break;
  case ModelType::eCexonPrc: width = 3; smP = nmrate; break;
  case ModelType::eAexonNpc: width = 1; smP = nmrate; break;
  case ModelType::eCexonNpc: width = 1; smP = nmrate; break;
  default: 
    cerr<<"no model used for: "<<ModelType::cname2str(mn)<<" "<<mn<<endl;
    break;
  } 
  if( width!=-1 ) {
    assert(width !=-1);
    assert(smP !=-1);
    const ParamFile* pf = ParamFile::getInstance();
    assert(pf);
    const Phylogeny& phy = pf->getPhylogeny();
    _pmSlow[mn] = new PhyloModel(phy,fmrate,width);
    _pmFast[mn] = new PhyloModel(phy,nmrate,width);
  }
}

void
InitNucDistr::train(const fasta& aligned_targ) {
  const ParamFile* pf = ParamFile::getInstance();
  const fasta& targ = pf->getTarg();
  const AnnotationInfo* ai = AnnotationInfo::getInstance();
  fasta::const_iterator iter = aligned_targ.begin();
  fasta::const_iterator stop = aligned_targ.end();
  for(; iter != stop; ++iter) {
    const fasta_seq& aligned_fs = **iter;
    const string& header = aligned_fs.get_header();
    fasta::const_iterator targ_match = targ.find_header(aligned_fs.get_header());
    if( targ_match == targ.end()) {
      cout<<"error what happend no alignment? ";
      cout<<aligned_fs.get_header()<<endl;
      continue;
    }
    const fasta_seq& targ_fs = **targ_match;
    //cout<<"Preliminary Training on: "<<targ_fs.get_header()<<endl;
    ParseHeader ph(ai);
    bool status = ph.parse(targ_fs);
    if(!status) {
      continue;
    }
    ExonDescrip& ed = ph.getExonDescrip();
    //cout<<"before ed: "<<endl;
    //ed.print();
    //cout<<endl;
    //const string& str = aligned_fs.get_seq();
    list<SeqLabel> klst = ed.getSeqLabelLst();
    const unsigned size = klst.size();
    assert(size >= 2 );
    list<SeqLabel>::const_iterator kiter = klst.begin();
    const list<SeqLabel>::const_iterator kstop = klst.end();
    vector<const SeqLabel*> lvec(klst.size(),NULL);
    for(unsigned i =0; kiter != kstop; ++kiter, ++i) {
      const SeqLabel& rght = *kiter;
      lvec[i] = &rght;
    }
    sort(lvec.begin(),lvec.end(),TCmp());
    const SeqLabel* left = lvec[0];
    const SeqLabel* pleft = NULL, *prght = NULL;

    const vector<string*> svec = getMatchSeq(aligned_fs);
    for(unsigned k = 1; k < lvec.size(); ++k) {
      const SeqLabel* rght = lvec[k];
      try 
	{
	  updateModel(svec,*left,*rght,ed);
	} 
      catch(const runtime_error& err) 
	{
	  cerr<<err.what()<<endl<<header<<endl;
	  break;
	}
      pleft = left;
      prght = rght;
      left = rght;
    }
  }
  //normalize();
  for(unsigned i = 0; i < (unsigned)ModelType::eCnameEnd; ++i) {
    cname_t mn = (cname_t)i;
    createPhyloModel(mn,1.6);
  }
}
