#include "ModelManager.h"
#include "ModelHash.h"
#include "MarkovChain.h"
#include "IhMc.h"
#include "ParseHeader.h"
#include "BaumWelch.h"
#include <vector>
#include <list>
#include <fstream>
#include <stdlib.h>
#include <math.h>
#include <algorithm>

using std::pair;
using std::cerr;
using std::endl;
using std::cout;
using std::string;
using std::vector;
using std::list;
using std::ofstream;
using std::ifstream;

bool SPECIAL_DEBUG=false;
bool GDEBUG=false;
extern bool DEBUGON;
static bool DEBUG=false;
bool XDEBUGON=false;
bool YDEBUGON=false;

static bool 
isConsCodonCheck(unsigned pos, int frame, const vector<string*>& svec); 
static void
updateHelp(ModelType::mname_t mn, const vector<string*>& svec, MultiModelHashT& mm, int pos, int relPos, bool isAlt); 

ModelManager::ModelManager(bool sep) : _rm(NULL), _sepNonCons(sep) {
  _stateTrans = new StateTrans;
  _rm = new RateManager;
  _rm->create();
  _cm = new ConsManager();
  _pf = ParamFile::getInstance();
  assert(_pf);
  for(unsigned i = 0; i < (unsigned)ModelType::eError; ++i) {
    ModelType::mname_t mn = (ModelType::mname_t)i;
    createModel(mn);
    createModelKmer(mn);
    createPhyloModel(mn);
  }
  for(unsigned i = 0; i < ModelType::eError ; ++i) {
    ModelType::mname_t id = (ModelType::mname_t)i;
    if( id != ModelType::eTerm ) {
      assert(_mm[id]);
    }
  }
}

double ModelManager::scoreTrans(const SeqLabel& left, const SeqLabel& rght) const { 
  const double val = _stateTrans->getProb(left,rght); 
  return val;
}

double ModelManager::scoreTrans(const SeqLabel& pleft, const SeqLabel& prght, const SeqLabel& cleft, const SeqLabel& crght) const {
  const double val = _stateTrans->getProb(pleft,prght,cleft,crght); 
  return val;
}

static void
dHelp(MultiModelHashT::iterator iter, const MultiModelHashT::iterator& end) {
  for(; iter != end; ++iter) {
    ModelType::mname_t id = (*iter).first;
    assert(id != ModelType::eError);
    if( id == ModelType::eTerm ) {
      continue;
    }
    MultiModel* md = (*iter).second;
    delete md;
  }
}

static void
dHelp(ModelHashT::iterator iter, const ModelHashT::iterator& end) {
  for(; iter != end; ++iter) {
    ModelType::mname_t id = (*iter).first;
    assert(id != ModelType::eError);
    if( id == ModelType::eTerm ) {
      continue;
    }
    Model* md = (*iter).second;
    delete md;
  }
}

static void
dPhyloHelp(ModelManager::PhyloModelHashT::iterator iter, const ModelManager::PhyloModelHashT::iterator& end) {
  for(; iter != end; ++iter) {
    ModelType::mname_t id = (*iter).first;
    assert(id != ModelType::eError);
    if( id == ModelType::eTerm ) {
      continue;
    }
    PhyloModel* md = (*iter).second;
    delete md;
  }
}


ModelManager::~ModelManager() {
  delete _stateTrans;
  delete _cm;
  MultiModelHashT::iterator iter = _mm.begin();
  const MultiModelHashT::iterator end = _mm.end();
  dHelp(iter,end);
  ModelHashT::iterator iter2 = _kmer.begin();
  const ModelHashT::iterator end2 = _kmer.end();
  dHelp(iter2,end2);

  PhyloModelHashT::iterator iter3 = _pm.begin();
  const PhyloModelHashT::iterator end3 = _pm.end();
  dPhyloHelp(iter3,end3);
}


#if 0
static 
void pHelp(ModelHashT::iterator iter, const ModelHashT::iterator& end) {
  for(; iter != end; ++iter) {
    ModelType::mname_t id = (*iter).first;
    assert(id != ModelType::eError);
    if( id == ModelType::eTerm ) {
      continue;
    }
    Model* md = (*iter).second;
    cout<<"Before: "<<ModelType::mname2str(id)<<endl;
    md->print(cout); cout<<endl;
    md->normalize();
    cout<<"After: "<<ModelType::mname2str(id)<<endl;
    md->print(cout); cout<<endl;
  }
}
#endif

static 
void qPrn(MultiModelHashT::iterator iter, const MultiModelHashT::iterator& end) {
  for(; iter != end; ++iter) {
    ModelType::mname_t id = (*iter).first;
    assert(id != ModelType::eError);
    if( id == ModelType::eTerm ) {
      continue;
    }
    MultiModel* md = (*iter).second;
    cout<<"Raw: "<<ModelType::mname2str(id)<<endl;
    md->print(cout); cout<<endl;
  }
}

void 
ModelManager::normalize() {
  //_cm->normalize();
  MultiModelHashT::iterator iter = _mm.begin();
  const MultiModelHashT::iterator end = _mm.end();
  //cout<<"Mc\n";
  qPrn(iter,end);
  //ModelHashT::iterator iter = _kmer.begin();
  //const ModelHashT::iterator end2 = _kmer.end();
  //cout<<"Kmers\n";
  //pHelp(iter,end2);
  //_stateTrans.hardCodeVals();
  _stateTrans->print(cout);
  _stateTrans->normalize();
}

void
ModelManager::write(const string& file) const {
  ofstream ofs(file.c_str()); 
  if( !ofs ) {
    cerr<<"Could not open ["<<file<<":"<<endl;
    assert(0);
  }
  MultiModelHashT::const_iterator iter = _mm.begin();
  const MultiModelHashT::const_iterator end = _mm.end();
  for(; iter != end; ++iter) {
    ModelType::mname_t id = (*iter).first;
    const MultiModel* md = (*iter).second;
    if( md ) {
      ofs<<ModelType::mname2str(id)<<endl;
      md->write(ofs);
    }
  }
  const string tstr = file + ".trans";
  _stateTrans->write(tstr);
  _stateTrans->print(cout);
  if( _rm ) {
     const string str4 = file + ".lfr";
     _rm->write(str4);
  }
}

bool
ModelManager::readLFR(const string& file) {
  const string str4 = file + ".lfr";
  return _rm->read(str4);
}

bool
ModelManager::read(const string& file) {
  ifstream ifs(file.c_str()); 
  if( !ifs ) {
    cerr<<"Could not open ["<<file<<"]"<<endl;
    return false;
  }
  string id;
  while(ifs>>id) {
    ModelType::mname_t mn = ModelType::str2mname(id);
    if(mn != ModelType::eError && mn != ModelType::eTerm) {
      _mm[mn]->read(ifs);
    }
  }
  const string tstr = file + ".trans";
  _stateTrans->read(tstr);
  //_stateTrans.drawGraphViz();
  const string str4 = file + ".lfr";
  if( !_rm->read(str4) ) {
  }
  return true;
}

std::pair<ModelType::mname_t,int>
ModelManager::getModel(const SeqLabel& leftLabel, const SeqLabel& rghtLabel, const ExonDescrip& exonDescrip, 
		       unsigned pos, const std::vector<string*>& svec, bool notStrict, int train_stage, bool& flag) {
  assert((signed)pos > leftLabel.getRght());
  ModelType::mname_t mn = ModelType::eError;
  const SeqLabel::blabel_t rbl = rghtLabel.getLabel();
  const SeqLabel::event_t rev = rghtLabel.getEvent();
  int consTypeRange = 2;

  if( rghtLabel.isIn(pos) ) {
    ModelType::cname_t distr = ModelType::sl2cnamePos(rghtLabel);
    if( distr != ModelType::eCnameEnd ) {
      int relPos = pos-rghtLabel.getLeft();
      assert(relPos>=0);
      const ipair_t consType = getSpliceConsType(pos,svec,distr,relPos,train_stage,flag);
      consTypeRange = consType.second;
      mn = ModelType::toSplice(rev,rbl,consType.first);
    }
  } else {
    // exon
    bool isExon = false;
    if( (SeqLabel::isAcc(leftLabel.getLabel()) || leftLabel.getLabel() == SeqLabel::eBeg) && 
	(SeqLabel::isDon(rghtLabel.getLabel()) || rghtLabel.getLabel() == SeqLabel::eEnd) ) {
      isExon = true;
    }
    // upstream intron
    else if( leftLabel.getLabel() == SeqLabel::eBeg && 
	     SeqLabel::isAcc(rghtLabel.getLabel()) ) {
      const ModelType::cname_t distr = ModelType::sl2cnameInv(leftLabel,rghtLabel,false);
      const ipair_t consType = getConsType(pos,svec,distr,exonDescrip,leftLabel,rghtLabel,train_stage,flag);
      consTypeRange = consType.second;
      mn = ModelType::toUpstreamIntron(rev,consType.first,notStrict);
    }
    // downstream intron
    else if( SeqLabel::isDon(leftLabel.getLabel()) &&
	     rghtLabel.getLabel() == SeqLabel::eEnd ) {
      const ModelType::cname_t distr = ModelType::sl2cnameInv(leftLabel,rghtLabel,false);
      const ipair_t consType = getConsType(pos,svec,distr,exonDescrip,leftLabel,rghtLabel,train_stage,flag);
      consTypeRange = consType.second;
      mn = ModelType::toDnstreamIntron(rev,consType.first,notStrict);
    }
    // retained intron
    else if( SeqLabel::isDon(leftLabel.getLabel()) &&
	     SeqLabel::isAcc(rghtLabel.getLabel())) {
      isExon = true;
    }
    // multiple donor region
    else if( SeqLabel::isDon(leftLabel.getLabel()) &&
	     SeqLabel::isDon(rghtLabel.getLabel()) ) {
      isExon = true;
    }
    // multiple acceptor region
    else if( SeqLabel::isAcc(leftLabel.getLabel()) &&
	     SeqLabel::isAcc(rghtLabel.getLabel()) ) {
      isExon = true;
    }
    if(isExon) {
      const int rpos = exonDescrip.getTargPos(pos);
      const bool isPrc = exonDescrip.isProteinCoding(rpos,leftLabel,rghtLabel);
      const ModelType::cname_t distr = ModelType::sl2cnameInv(leftLabel,rghtLabel,isPrc);
      if( distr != ModelType::eCnameEnd ) {
	const ipair_t consType = getConsType(pos,svec,distr,exonDescrip,leftLabel,rghtLabel,train_stage,flag);
	consTypeRange = consType.second;
	mn = ModelType::toExon(rev,consType.first,isPrc,notStrict);
      }
    }
  }
  return pair<ModelType::mname_t,int>(mn,consTypeRange);
}

static int
getConsType(dpair_t dp, double score1) {
  int consType = 2;
  double diff = dp.first-dp.second;
  diff /= 2;
  if( score1 >= dp.first ) {
    consType = 0; //ModelType::eConserved;
  } else if( score1 < dp.first && score1 >= (dp.first-diff)) {
    consType = 1;
  } else if( score1 < (dp.first-diff) && score1 >= (dp.first-(2*diff))) {
    consType = 2;
  } else if( score1 <= dp.second ) {
    consType = 3; //ModelType::eNotConserved;
  }
  return consType;
}

ipair_t
ModelManager::getSpliceConsType(unsigned i, const std::vector<string*>& svec, cname_t typ, int relPos, int train_stage, bool& flag) {
  if( svec.size()== 1) {
    return ipair_t(0,0);
  }
  int consType = 2, consTypeRange = 2;
  const unsigned half = 0;
  if(i >= half && (i+half < svec[0]->size()) ) { 
    consType = 0;
    consTypeRange = 0;
  }
  return ipair_t(consType,consTypeRange);
}

int
ModelManager::getIntronConsLevel(unsigned i, const std::vector<string*>& svec, cname_t typ) {
  int consType = 3;
  const unsigned half = 5;
  if(i >= half && ((i+half) < svec[0]->size()) ) { 
    const ipair_t bnd( (i-half)+1,(i+half)+1);
    double score1 = 0;
    const int frame = 0;
    unsigned num_org = svec.size();
    for(signed pos = bnd.first; pos <= bnd.second; ++pos) {
      string column(num_org,'-');
      fillColumn(column,svec,pos);
      ModelType::mname_t con_mn = ModelType::cname2mname(typ,0);
      PhyloModel* cm = _pm[con_mn];
      MultiModel* cmcm = _mm[con_mn];
      double cprob = cm->getProb(svec,pos,frame,column,cmcm);
      ModelType::mname_t non_mn = ModelType::cname2mname(typ,3);
      PhyloModel* nm = _pm[non_mn];
      MultiModel* nmcm = _mm[non_mn];
      double nprob = nm->getProb(svec,pos,frame,column,nmcm);
      if( cprob > 0 && nprob > 0 ) {
	score1 += (log2(cprob)-log2(nprob));
      }
    }
    dpair_t dp = _cthresh.getConsThresh(typ);
    consType = ::getConsType(dp,score1);
  }
  return consType;
}

ipair_t
ModelManager::getConsType(unsigned i, const std::vector<string*>& svec, cname_t typ, const ExonDescrip& exonDescrip,
			  const SeqLabel& leftLabel, const SeqLabel& rghtLabel, int train_stage, bool& flag) {
  if( svec.size()== 1) {
    return ipair_t(0,0);
  }
  assert( !ModelType::isSplice(typ) );
  int consType = 0, consTypeRange = 0;
  if( _sepNonCons && ModelType::isIntron(typ) ) {
    bool isConsVal=false;
    if( ModelType::isPrcExon(typ) ) {
      int frame = exonDescrip.getFrame(i,leftLabel,rghtLabel);
      isConsVal = isConsCodonCheck(i,frame,svec);
    } else {
      isConsVal = isCons(i,svec,-1);
    }
    if( isConsVal ) {
      consType = 0;
    } else {
      consType = 3;
    }
  }
  return ipair_t(consType,consTypeRange);
}

bool
ModelManager::updateModel(const string& str, const SeqLabel& leftLabel, const SeqLabel& rghtLabel,
			  const ExonDescrip& targetCoords,
			  const std::vector<string*>& svec, bool notStrict, int train_stage) {
  bool flag_raise = false;
  for(signed pos = leftLabel.getRght()+1; pos <= rghtLabel.getRght(); ++pos) {
    pair<ModelType::mname_t,int> pr = getModel(leftLabel,rghtLabel,targetCoords,pos,svec,notStrict,train_stage,flag_raise);
    if(train_stage==0)
      continue;
    ModelType::mname_t mn = pr.first;
    if( mn != ModelType::eError ) {
      int relPos = -1;
      bool fail = false;
      bool ldebug=false;
      int frame=-1;
      if(DEBUG) cout<<"slabel "<<pos<<": "<<ModelType::mname2str(mn)<<endl;
      if( rghtLabel.isIn(pos) ) {
	if( (signed)SeqLabel::getSigWidth(rghtLabel.getLabel()) != rghtLabel.getLength()) {
	  cout<<"skip, this splice site includes insertions: "<<rghtLabel.getLength()<<" "<<rghtLabel<<endl;
	  continue;
	} else {
	  //cout<<"don't skip here: "<<rghtLabel<<endl;
	}
	// get position in signal
	relPos = pos-rghtLabel.getLeft() ;
	const int left = rghtLabel.getLeft();
	char c1='x',c2='x';
	const unsigned idx = (left+relPos)-1;
	int tRelPos = 20;
	if( tRelPos == 20 && SeqLabel::isAcc(rghtLabel.getLabel())) {
	  c1 = idx < str.size() ? tolower(str[left+tRelPos-1]) : 'x';
	  if(c1!='a') {
	    fail=true;
	  }
	}
	tRelPos = 21;
	if( tRelPos == 21 && SeqLabel::isAcc(rghtLabel.getLabel())) {
	  c2 = idx < str.size() ? tolower(str[left+tRelPos-1]) : 'x';
	  if( c2 !='g') {
	    fail=true;
	  }
	}
	tRelPos = 3;
	if( tRelPos == 3 && SeqLabel::isDon(rghtLabel.getLabel())) {
	  c1 = idx < str.size() ? tolower(str[left+tRelPos-1]) : 'x';
	  if( c1 != 'g') 
	    fail=true;
	  ldebug=true;
	}
	tRelPos = 4;
	if( tRelPos == 4 && SeqLabel::isDon(rghtLabel.getLabel())) {
	  c2 = idx < str.size() ? tolower(str[left+tRelPos-1]) : 'x';
	  if( c2 != 't' && c2 != 'c') 
	    fail=true;
	  ldebug=true;
	}
	if( fail ) {
	  string sub = str.substr(left-1,(rghtLabel.getRght()-left)+1);
	  //cout<<"funky splice site hmm: ["<<c1<<"] ["<<c2<<"] "<<sub<<endl;
	  //cout<<"coords: "<<left<<" "<<relPos<<" "<<rghtLabel<<endl;
	  flag_raise = true;
	}
      }
      else if(  ModelType::isPrcExon(mn) ) {
	const int rpos = targetCoords.getTargPos(pos);
	assert(rpos != -1);
	if(DEBUG)
	  cout<<"orig: "<<pos<<" rpos: "<<rpos<<endl;
	frame=targetCoords.getFrame(rpos,leftLabel,rghtLabel);
	assert(frame != -1);
	//cout<<"frame: "<<frame<<endl;
	relPos=frame;
	if( pos-3 >= 0) {
	  string codon = str.substr(pos-3,3);
	  makelower(codon);
	  if(DEBUG) {
	    cout<<codon<<" frame: "<<frame<<endl;
	  }
	  if(frame == 2) {
	    if( codon=="taa" || codon=="tga"||codon=="tag") {
	      ExonDescrip::ipair_t ip = targetCoords.getStartStop();	
	      cout<<"start/stop: "<<ip.first<<" "<<ip.second<<endl;
	      cout<<"stop codon: "<<codon<<" "<<leftLabel<<" "<<rghtLabel<<" pos: "<<pos<<endl;
	      //const int tmp=targetCoords.getFrame(rpos,leftLabel,rghtLabel);
	      //cout<<"stop codon quit"<<endl;
	      break;
	    }
	  }
	}
      } else {
	relPos = pos+1;
      }
      assert(relPos != -1);
      if( !fail &&  mn != ModelType::eTerm ) {
	const bool isAlt = rghtLabel.getEvent()==SeqLabel::eAlt;
#if 0
	const cname_t cn = ModelType::mname2cname(mn,isAlt);
	_mm[mn]->update(svec,pos,relPos);
	if( ModelType::isIntron(cn) ) {
	  const ModelType::mname_t tmn = ModelType::getOppSide(mn);
	  _mm[tmn]->update(svec,pos,relPos);
	  if(ModelType::isNonCons(tmn)) {
	    const ModelType::mname_t tmn1 = ModelType::getOppCons(tmn);
	    _mm[tmn1]->update(svec,pos,relPos);
	  }
	}
	if( ModelType::isSplice(cn) ) {
	  const ModelType::cname_t ocn = ModelType::getOppSplice(cn);
	  const ModelType::mname_t omn = ModelType::cname2mname(ocn,0);
	  _mm[omn]->update(svec,pos,relPos);
	}
	if( ModelType::isNonCons(mn) ) {
	  const ModelType::mname_t tmn = ModelType::getOppCons(mn);
	  _mm[tmn]->update(svec,pos,relPos);
	}

	Model* kmer = _kmer[mn];
	kmer->update(str,pos,relPos);
#endif
	updateHelp(mn,svec,_mm,pos,relPos,isAlt);
      }
    }
  }
  return flag_raise;
}

void
updateHelp(ModelType::mname_t mn, const vector<string*>& svec, MultiModelHashT& mm, int pos, int relPos, bool isAlt) {
  //manually tied
  const ModelType::cname_t cn = ModelType::mname2cname(mn,isAlt);
  if(ModelType::isIntron(cn)) {
    if(isAlt) {
      mm[ModelType::eAintronNonConDn]->update(svec,pos,relPos);
      mm[ModelType::eAintronConDn]->update(svec,pos,relPos);
      mm[ModelType::eAintronNonConUp]->update(svec,pos,relPos);
      mm[ModelType::eAintronConUp]->update(svec,pos,relPos);

      mm[ModelType::eCintronNonConDn]->update1(svec,pos,relPos);
      mm[ModelType::eCintronConDn]->update1(svec,pos,relPos);
      mm[ModelType::eCintronNonConUp]->update1(svec,pos,relPos);
      mm[ModelType::eCintronConUp]->update1(svec,pos,relPos);

    } else {
      mm[ModelType::eCintronNonConDn]->update(svec,pos,relPos);
      mm[ModelType::eCintronConDn]->update(svec,pos,relPos);
      mm[ModelType::eCintronNonConUp]->update(svec,pos,relPos);
      mm[ModelType::eCintronConUp]->update(svec,pos,relPos);

      mm[ModelType::eAintronNonConDn]->update1(svec,pos,relPos);
      mm[ModelType::eAintronConDn]->update1(svec,pos,relPos);
      mm[ModelType::eAintronNonConUp]->update1(svec,pos,relPos);
      mm[ModelType::eAintronConUp]->update1(svec,pos,relPos);
    }
  } else if(ModelType::isSplice(cn)) {
    mm[mn]->update(svec,pos,relPos);
    const ModelType::cname_t ocn = ModelType::getOppSplice(cn);
    const ModelType::mname_t omn = ModelType::cname2mname(ocn,0);
    mm[omn]->update(svec,pos,relPos);
  } else if(ModelType::isPrcExon(cn)) {
    if(isAlt) {
      mm[ModelType::eAexonNonConPrc]->update(svec,pos,relPos);
      mm[ModelType::eAexonConPrc]->update(svec,pos,relPos);

      mm[ModelType::eCexonNonConPrc]->update1(svec,pos,relPos);
      mm[ModelType::eCexonConPrc]->update1(svec,pos,relPos);
    } else {
      mm[ModelType::eCexonNonConPrc]->update(svec,pos,relPos);
      mm[ModelType::eCexonConPrc]->update(svec,pos,relPos);

      mm[ModelType::eAexonNonConPrc]->update1(svec,pos,relPos);
      mm[ModelType::eAexonConPrc]->update1(svec,pos,relPos);
    }
  } else if(ModelType::isNpcExon(cn)) {
    if(isAlt) {
      mm[ModelType::eAexonNonConNpc]->update(svec,pos,relPos);
      mm[ModelType::eAexonConNpc]->update(svec,pos,relPos);

      mm[ModelType::eCexonNonConNpc]->update1(svec,pos,relPos);
      mm[ModelType::eCexonConNpc]->update1(svec,pos,relPos);
    } else {
      mm[ModelType::eCexonNonConNpc]->update(svec,pos,relPos);
      mm[ModelType::eCexonConNpc]->update(svec,pos,relPos);

      mm[ModelType::eAexonNonConNpc]->update1(svec,pos,relPos);
      mm[ModelType::eAexonConNpc]->update1(svec,pos,relPos);
    }
  } else {
    assert(0);
  }
}


void
ModelManager::updateFalseModel(const string& str, const SeqLabel& leftLabel, const SeqLabel& rghtLabel,
			  const ExonDescrip& targetCoords,
			  const std::vector<string*>& svec, bool notStrict) {
  signed slen =(signed) str.length();
  unsigned donCnt =0;
  list<int> lst;
  for(signed pos = leftLabel.getRght()+1; pos < rghtLabel.getLeft(); ++pos) {
    const ipair_t dp = SeqLabel::getSigDim(SeqLabel::eDon,pos);
    const char a = tolower(str[pos]);
    const char b = tolower(str[pos+1]);
    if( a == 'g' && (b=='t'||b=='c') &&
	dp.first >= 1 && dp.second <= slen ) {
      if( b=='c' ) {
	lst.push_back(pos);
      }
      ++donCnt;
    }
  }
  vector<int> rpos(lst.size());
  list<int>::const_iterator iter = lst.begin();
  const list<int>::const_iterator stop = lst.end();
  for(unsigned idx = 0; iter != stop; ++idx, ++iter) {
    rpos[idx]=*iter;
  }
  std::random_shuffle(rpos.begin(),rpos.end());
  //double gcPcnt = ((double)dc.first/(double)dc.second) * 100;
  double gcPcnt=0;
  for(unsigned i = 0; (double)i < gcPcnt && i < rpos.size(); ++i) {
    const int pos = rpos[i];
    const ipair_t dp = SeqLabel::getSigDim(SeqLabel::eDon,pos);
    MultiModel* modc = _mm[ModelType::eDonFalseCons];
    MultiModel* modn = _mm[ModelType::eDonFalseNonc];
    for(signed lpos = dp.first, rpos = 0; lpos <= dp.second; ++lpos, ++rpos) {
      modc->update(svec,lpos,rpos);
      modn->update(svec,lpos,rpos);
    }
  }
  for(signed pos = leftLabel.getRght()+1; pos < rghtLabel.getLeft(); ++pos) {
    MultiModel* modn = _mm[ModelType::eBackgroundCons];
    MultiModel* modc = _mm[ModelType::eBackgroundNonc];
    modn->update(svec,pos,pos+1);
    modc->update(svec,pos,pos+1);
    ipair_t ap = SeqLabel::getSigDim(SeqLabel::eAcc,pos);
    ipair_t dp = SeqLabel::getSigDim(SeqLabel::eDon,pos);
    const char a1 = tolower(str[pos-3]);
    const char b1 = tolower(str[pos-2]);
    if( a1 == 'a' && b1=='g' &&
	ap.first >= 1 && ap.second <= slen ) {
      MultiModel* modc = _mm[ModelType::eAccFalseCons];
      MultiModel* modn = _mm[ModelType::eAccFalseNonc];
      for(signed lpos = ap.first, rpos = 0; lpos <= ap.second; ++lpos, ++rpos) {
	modn->update(svec,lpos,rpos);
	modc->update(svec,lpos,rpos);
      }
    }
    const char a = tolower(str[pos]);
    const char b = tolower(str[pos+1]);
    if( a == 'g' && b=='t' &&
	dp.first >= 1 && dp.second <= slen ) {
      MultiModel* modc = _mm[ModelType::eDonFalseCons];
      MultiModel* modn = _mm[ModelType::eDonFalseNonc];
      for(signed lpos = dp.first, rpos = 0; lpos <= dp.second; ++lpos, ++rpos) {
	modc->update(svec,lpos,rpos);
	modn->update(svec,lpos,rpos);
      }
    }
  }
}


struct TCmp {
  bool operator()(const SeqLabel* a, const SeqLabel* b) const {
    return a->getLeft()<b->getLeft();
  }
};

void
ModelManager::train(const fasta& aligned_targ) {
  trainStage(aligned_targ,1);
  trainFalse(aligned_targ);
  normalize();
}

void
ModelManager::trainRates(const fasta& aligned_targ) {
  //const ParamFile* pf = ParamFile::getInstance();
  //const fasta& targ = pf->getTarg();
  const AnnotationInfo* ai = AnnotationInfo::getInstance();
  for(unsigned cnt = 0; cnt < 30; ++cnt) {
    vector<double> prob_sum(ModelType::eError,0.0);
    vector<double> prob_cnt(ModelType::eError,0.0);
    cout<<"BM iteration "<<cnt<<endl;
    fasta::const_iterator iter = aligned_targ.begin();
    const fasta::const_iterator stop = aligned_targ.end();
    for(; iter != stop; ++iter) {
#if 0
      const fasta_seq& aligned_fs = **iter;
      fasta::const_iterator targ_match = targ.find_header(aligned_fs.get_header());
      if( targ_match == targ.end()) {
	cout<<"error what happend no alignment? "<<aligned_fs.get_header()<<endl;
	continue;
      }
#endif
      const fasta_seq& targ_fs = **iter; //**targ_match;
      const vector<string*> svec = getMatchSeq(targ_fs);
      ParseHeader ph(ai);
      bool status = ph.parse(targ_fs);
      if(!status) {
	continue;
      }
      //const string& head = aligned_fs.get_header();
      //if(head != 
      //"CG14691_3R:-_CLASS:_MS_[RA,6579439_6579598:RB,6579439_6579561],_[RA,6579773_6579887:RB,6579773_6579847]_,[RA,6579950_6580106:RB,6579909_6580106]") {
       //continue;
      //}
      cout<<targ_fs.get_header()<<endl; 
      ExonDescrip& ed = ph.getExonDescrip();
      ed.print(cout);
      //const string& str = aligned_fs.get_seq();
      list<SeqLabel> klst = ed.getSeqLabelLst();
      const unsigned size = klst.size();
      assert(size >= 2 );
      list<SeqLabel>::const_iterator kiter = klst.begin();
      const list<SeqLabel>::const_iterator kstop = klst.end();
      vector<const SeqLabel*> lvec(klst.size(),NULL);
      for(unsigned i =0; kiter != kstop; ++kiter, ++i) {
	const SeqLabel& rght = *kiter;
	lvec[i] = &rght;
      }
      sort(lvec.begin(),lvec.end(),TCmp());
      const SeqLabel* left = lvec[0];
      const SeqLabel* pleft = NULL, *prght = NULL;
      bool flag = false;
      for(unsigned k = 1; k < lvec.size(); ++k) {
	const SeqLabel* rght = lvec[k];
	runBM(*left,*rght,ed,svec,prob_sum,prob_cnt);
	if( status ) {
	  flag = true;
	}
	pleft = left;
	prght = rght;
	left = rght;
      }
      
    }
    cout<<"The averages begin: "<<endl;
    for(unsigned i = 0; i < prob_sum.size(); ++i) {
      if(prob_cnt[i] > 0 ) {
	const double val = prob_sum[i]/prob_cnt[i];
	mname_t mn = (mname_t)i;
	cout<<ModelType::mname2str(mn)<<" avg "<<cnt<<" "<<val<<endl;
      }
    }
    cout<<"The averages end: "<<endl;
    _cm->normalize();
  }
}

void
ModelManager::trainFalse(const fasta& aligned_targ) {
  //const ParamFile* pf = ParamFile::getInstance();
  //const fasta& targ = pf->getTarg();
  const AnnotationInfo* ai = AnnotationInfo::getInstance();
  fasta::const_iterator iter = aligned_targ.begin();
  const fasta::const_iterator stop = aligned_targ.end();
  for(; iter != stop; ++iter) {
#if 0
    const fasta_seq& aligned_fs = **iter;
    fasta::const_iterator targ_match = targ.find_header(aligned_fs.get_header());
    if( targ_match == targ.end()) {
      cout<<"error what happend no alignment? ";
      cout<<aligned_fs.get_header()<<endl;
      continue;
    }
#endif
    const fasta_seq& targ_fs = **iter; //**targ_match;
    const vector<string*> svec = getMatchSeq(targ_fs);
    ParseHeader ph(ai);
    bool status = ph.parse(targ_fs);
    if(!status) {
      continue;
    }
    ExonDescrip& ed = ph.getExonDescrip();
    const string& str = targ_fs.get_seq();
    list<SeqLabel> klst = ed.getSeqLabelLst();
    const unsigned size = klst.size();
    assert(size >= 2 );
    list<SeqLabel>::const_iterator kiter = klst.begin();
    const list<SeqLabel>::const_iterator kstop = klst.end();
    vector<const SeqLabel*> lvec(klst.size(),NULL);
    for(unsigned i =0; kiter != kstop; ++kiter, ++i) {
      const SeqLabel& rght = *kiter;
      lvec[i] = &rght;
    }
    sort(lvec.begin(),lvec.end(),TCmp());
    const SeqLabel* left = lvec[0];
    const SeqLabel* pleft = NULL, *prght = NULL;
    for(unsigned k = 1; k < lvec.size(); ++k) {
      const SeqLabel* rght = lvec[k];
      updateFalseModel(str,*left,*rght,ed,svec,false);
      pleft = left;
      prght = rght;
      left = rght;
    }
  }
}

void
ModelManager::trainStage(const fasta& aligned_targ, int train_stage) {
  //const ParamFile* pf = ParamFile::getInstance();
  //const fasta& targ = pf->getTarg();
  const AnnotationInfo* ai = AnnotationInfo::getInstance();
  fasta::const_iterator iter = aligned_targ.begin();
  fasta::const_iterator stop = aligned_targ.end();
  for(; iter != stop; ++iter) {
#if 0
    const fasta_seq& aligned_fs = **iter;
    fasta::const_iterator targ_match = targ.find_header(aligned_fs.get_header());
    if( targ_match == targ.end()) {
      cout<<"error what happend no alignment? ";
      cout<<aligned_fs.get_header()<<endl;
      continue;
    }
#endif
    const fasta_seq& targ_fs = **iter; //**targ_match;
    const vector<string*> svec = getMatchSeq(targ_fs);
    ParseHeader ph(ai);
    bool status = ph.parse(targ_fs);
    if(!status) {
      continue;
    }
    ExonDescrip& ed = ph.getExonDescrip();
    const string& str = targ_fs.get_seq();
    list<SeqLabel> klst = ed.getSeqLabelLst();
    const unsigned size = klst.size();
    assert(size >= 2 );
    list<SeqLabel>::const_iterator kiter = klst.begin();
    const list<SeqLabel>::const_iterator kstop = klst.end();
    vector<const SeqLabel*> lvec(klst.size(),NULL);
    for(unsigned i =0; kiter != kstop; ++kiter, ++i) {
      const SeqLabel& rght = *kiter;
      lvec[i] = &rght;
    }
    sort(lvec.begin(),lvec.end(),TCmp());
    const SeqLabel* left = lvec[0];
    const SeqLabel* pleft = NULL, *prght = NULL;
    bool flag = false;
    for(unsigned k = 1; k < lvec.size(); ++k) {
      const SeqLabel* rght = lvec[k];
      bool status = updateModel(str,*left,*rght,ed,svec,false,train_stage);
      if( status ) {
	flag = true;
      }
      if( train_stage > 0 ) {
	if( k == 1 ) {
	  _stateTrans->update(*left,*rght);
	} else {
	  _stateTrans->update(*pleft,*prght,*left,*rght);
	}
      }
      pleft = left;
      prght = rght;
      left = rght;
    }
  }
}

void
ModelManager::createModel(ModelType::mname_t mn) {
  const ParamFile* pf = ParamFile::getInstance();
  const OrgIdx& oidx = pf->getOrgIdx();
  const unsigned size = oidx.size();
  MultiModel* md = NULL;
  const MarkovChain* mc1 = new MarkovChain(1);
  const MarkovChain* mc2 = new MarkovChain(2);
  const MarkovChain* mc3 = new MarkovChain(3);
  switch(mn) {
  case ModelType::eAdonCon: md = new MultiModel(new IhMc(mc1,9,0.0), size); break;
  case ModelType::eCdonCon: md = new MultiModel(new IhMc(mc1,9,0.0), size); break;
  case ModelType::eAaccCon: md = new MultiModel(new IhMc(mc1,24,0.0), size); break;
  case ModelType::eCaccCon: md = new MultiModel(new IhMc(mc1,24,0.0), size); break;
  case ModelType::eDonNonc: md = new MultiModel(new IhMc(mc1,9,0.0), size); break;
  case ModelType::eAccNonc: md = new MultiModel(new IhMc(mc1,24,0.0), size); break;
  case ModelType::eDonFalseNonc: md = new MultiModel(new IhMc(mc1,9,0.0), size); break;
  case ModelType::eDonFalseCons: md = new MultiModel(new IhMc(mc1,9,0.0), size); break;
  case ModelType::eAccFalseNonc: md = new MultiModel(new IhMc(mc1,24,0.0), size); break;
  case ModelType::eAccFalseCons: md = new MultiModel(new IhMc(mc1,24,0.0), size); break;

  case ModelType::eCintronNonConDn: md = new MultiModel(new MarkovChain(3), size); break;
  case ModelType::eCintronConDn: md = new MultiModel(new MarkovChain(3), size); break;
  case ModelType::eAintronNonConDn: md = new MultiModel(new MarkovChain(3), size); break;
  case ModelType::eAintronConDn: md = new MultiModel(new MarkovChain(3), size); break;

  case ModelType::eCintronNonConUp: md = new MultiModel(new MarkovChain(3), size); break;
  case ModelType::eCintronConUp: md = new MultiModel(new MarkovChain(3), size); break;
  case ModelType::eAintronNonConUp: md = new MultiModel(new MarkovChain(3), size); break;
  case ModelType::eAintronConUp: md = new MultiModel(new MarkovChain(3), size); break;

    // 3-periodic non-homogenious 2nd order MC
  case ModelType::eCexonNonConPrc: md = new MultiModel(new IhMc(mc2,3,0.0), size); break;
  case ModelType::eCexonConPrc: md = new MultiModel(new IhMc(mc3,3,0.0), size); break;
  case ModelType::eAexonConPrc: md = new MultiModel(new IhMc(mc3,3,0.0), size); break;
  case ModelType::eAexonNonConPrc: md = new MultiModel(new IhMc(mc2,3,0.0), size); break;

  case ModelType::eCexonNonConNpc: md = new MultiModel(new MarkovChain(2), size); break;
  case ModelType::eCexonConNpc: md = new MultiModel(new MarkovChain(2), size); break;
  case ModelType::eAexonConNpc: md = new MultiModel(new MarkovChain(2), size); break;
  case ModelType::eAexonNonConNpc: md = new MultiModel(new MarkovChain(2), size); break;
  case ModelType::eBackgroundCons: md = new MultiModel(new MarkovChain(2), size); break;
  case ModelType::eBackgroundNonc: md = new MultiModel(new MarkovChain(2), size); break;
  default: 
    //cerr<<"no model used for: "<<mname2str(mn)<<" "<<mn<<endl;
    break;
  } 
  delete mc1;
  delete mc2;
  delete mc3;
  if( md ) {
    _mm[mn] = md;
  }
}

void
ModelManager::createModelKmer(ModelType::mname_t mn) {
  Model* md = NULL;
  const Kmer* km3 = new Kmer(2);
  switch(mn) {
  case ModelType::eAdonCon: md = new IhMc(km3,9,0.0); break;
  case ModelType::eCdonCon: md = new IhMc(km3,9,0.0); break;
  case ModelType::eAaccCon: md = new IhMc(km3,24,0.0); break;
  case ModelType::eCaccCon: md = new IhMc(km3,24,0.0); break;
  case ModelType::eDonNonc: md = new IhMc(km3,9,0.0); break;
  case ModelType::eAccNonc: md = new IhMc(km3,24,0.0); break;
  case ModelType::eDonFalseCons: md = new IhMc(km3,9,0.0); break;
  case ModelType::eDonFalseNonc: md = new IhMc(km3,9,0.0); break;
  case ModelType::eAccFalseCons: md = new IhMc(km3,24,0.0); break;
  case ModelType::eAccFalseNonc: md = new IhMc(km3,24,0.0); break;

  case ModelType::eCintronNonConDn: md = new Kmer(3); break;
  case ModelType::eCintronConDn: md = new Kmer(3); break;
  case ModelType::eAintronNonConDn: md = new Kmer(3); break;
  case ModelType::eAintronConDn: md = new Kmer(2); break;

  case ModelType::eCintronNonConUp: md = new Kmer(3); break;
  case ModelType::eCintronConUp: md = new Kmer(3); break;
  case ModelType::eAintronNonConUp: md = new Kmer(3); break;
  case ModelType::eAintronConUp: md = new Kmer(3); break;

    // 3-periodic non-homogenious 2nd order MC
  case ModelType::eCexonNonConPrc: md = new IhMc(km3,3,0.0); break;
  case ModelType::eCexonConPrc: md = new IhMc(km3,3,0.0); break;
  case ModelType::eAexonConPrc: md = new IhMc(km3,3,0.0); break;
  case ModelType::eAexonNonConPrc: md = new IhMc(km3,3,0.0); break;

  case ModelType::eCexonNonConNpc: md = new Kmer(2); break;
  case ModelType::eCexonConNpc: md = new Kmer(2); break;
  case ModelType::eAexonConNpc: md = new Kmer(2); break;
  case ModelType::eAexonNonConNpc: md = new Kmer(2); break;
  case ModelType::eBackgroundCons: md = new Kmer(2); break;
  case ModelType::eBackgroundNonc: md = new Kmer(2); break;
  default: 
    //cerr<<"no model used for: "<<mname2str(mn)<<" "<<mn<<endl;
    break;
  } 
  delete km3;
  if( md ) {
    _kmer[mn] = md;
  }
}

void
ModelManager::createPhyloModel(ModelType::mname_t mn) {
  int width = -1; 
  double smP = -1;
  bool useCache = true;
  const double fmrate = 1.0;
  const double nmrate = 1.6;
  const double nonConIntron = 1.7;
  const double consCodon = 0.7;
  const double consAcc = 1.1;
  switch(mn) {
  case ModelType::eAdonCon: width = 9; smP = 1.0; break;
  case ModelType::eCdonCon: width = 9; smP = 1.0; break;
  case ModelType::eAaccCon: width = 24; smP = consAcc; break;
  case ModelType::eCaccCon: width = 24; smP = consAcc;break;
  case ModelType::eDonNonc: width = 9; smP = fmrate;break;
  case ModelType::eAccNonc: width = 24; smP = fmrate;break;
  case ModelType::eDonFalseNonc: width = 9; smP = fmrate;  break;
  case ModelType::eDonFalseCons: width = 9; smP = 1 /*fmrate*/;break;
  case ModelType::eAccFalseNonc: width = 24; smP = fmrate;break;
  case ModelType::eAccFalseCons: width = 24; smP = 1.0 /*fmrate*/; break;

  case ModelType::eCintronNonConDn: width = 1; smP = nonConIntron; useCache = true; break;
  case ModelType::eCintronConDn: width = 1; smP = 1.0 /*fmrate*/; break;
  case ModelType::eAintronNonConDn: width = 1; smP = nonConIntron; useCache = true;  break;
  case ModelType::eAintronConDn: width = 1; smP = 1.0/*fmrate*/; break;

  case ModelType::eCintronNonConUp: width = 1; smP = nonConIntron; break;
  case ModelType::eCintronConUp: width = 1; smP = 1.0/*fmrate*/; useCache = true;  break;
  case ModelType::eAintronNonConUp: width = 1; smP = nonConIntron; break;
  case ModelType::eAintronConUp: width = 1; smP = 1.0/*fmrate*/; useCache = true;  break;

    // 3-periodic non-homogenious 2nd order MC
  case ModelType::eCexonNonConPrc: width = 3; smP = nmrate; break;
  case ModelType::eCexonConPrc: width = 3; smP = consCodon; break;
  case ModelType::eAexonConPrc: width = 3; smP = consCodon; break;
  case ModelType::eAexonNonConPrc: width = 3; smP = 1.7/*nmrate*/;break;

  case ModelType::eCexonNonConNpc: width = 1; smP = nmrate; break;
  case ModelType::eCexonConNpc: width = 1; smP = fmrate; break;
  case ModelType::eAexonConNpc:width = 1; smP = fmrate; break;
  case ModelType::eAexonNonConNpc: width = 1; smP = nmrate; break;
  case ModelType::eBackgroundNonc: width = 1; smP=nmrate; useCache = true;  break;
  case ModelType::eBackgroundCons: width = 1; smP=fmrate; useCache = true;  break;
  default: 
    //cerr<<"no model used for: "<<mname2str(mn)<<" "<<mn<<endl;
    break;
  } 
  if( width!=-1 ) {
    assert(width !=-1);
    assert(smP !=-1);
    _pf = ParamFile::getInstance();
    assert(_pf);
    const Phylogeny& phy = _pf->getPhylogeny();
    _pm[mn] = new PhyloModel(phy,smP,width);
  }
}

static int mod2frame(int mod) {
  switch(mod) {
  case 0: return 2;
  case 1: return 0;
  case 2: return 1;
  default:
    assert(0); return -1;
  }
}


bool
ModelManager::runBM(const SeqLabel& left, const SeqLabel& rght, const ExonDescrip& targetCoords,
		    const std::vector<string*>& svec, vector<double>& prob_sum, vector<double>& prob_cnt) {
  ipair_t ip = targetCoords.getStartStop();
  cout<<"start/stop: "<<ip<<endl;
  vector<int> ftrack(3);
  if( left.getRght() >= rght.getLeft() ) {
    return false;
  }
  if( left.getLabel() != SeqLabel::eBeg ) {
    assert(!left.isEnd());
    const cname_t cn = ModelType::sl2cnamePos(left);
    RateHmm* rm = _cm->getRateHmm(cn);
    ModelType::mname_t mn = ModelType::cname2mname(cn,0);
    PhyloModel* pm = _pm[mn];
    MultiModel* mm = _mm[mn];
    BaumWelch bm(*rm,*pm,*mm);
    const int beg = left.getLeft()>=4 ? left.getLeft() : 4;
    const int end = left.getRght();
    int off=0;
    if( left.getLeft() < 4 ) {
      off=4-left.getLeft();
    }
    const int bframe = off;
    const int eframe = end-beg;
    bm.train(svec,beg,end,bframe,eframe);
    const double prob = bm.getLastProb();
    prob_sum[mn] += prob;
    ++prob_cnt[mn];
    cout<<"tprob: "<<ModelType::mname2str(mn)<<" "<<prob<<endl;
  }
  const bool isExon = SeqLabel::isExon(left.getLabel(),rght.getLabel());
  const cname_t cnNpc = ModelType::sl2cnameInv(left,rght,false);
  if( cnNpc == ModelType::eCnameEnd ) {
    cout<<"unclassified case"<<left<<" "<<rght<<endl;
    return false;
  }
  RateHmm* rmNpc = _cm->getRateHmm(cnNpc);
  ModelType::mname_t mnNpc = ModelType::cname2mname(cnNpc,0);
  PhyloModel* pmNpc = _pm[mnNpc];
  MultiModel* mmNpc = _mm[mnNpc];
  const int bframeNpc = -1;
  const int eframeNpc = -1;
  const cname_t cnPrc = ModelType::sl2cnameInv(left,rght,true);
  RateHmm* rmPrc = _cm->getRateHmm(cnPrc);
  ModelType::mname_t mnPrc = ModelType::cname2mname(cnPrc,0);
  PhyloModel* pmPrc = _pm[mnPrc];
  MultiModel* mmPrc = _mm[mnPrc];
  BaumWelch bmPrc(*rmPrc,*pmPrc,*mmPrc);
  BaumWelch bmNpc(*rmNpc,*pmNpc,*mmNpc);

  if( isExon && (ip.first > (left.getRght()+1) && ip.first <= (rght.getLeft()-1)) ||
      (ip.second > (left.getRght()+1) && ip.second <= (rght.getLeft()-1)) )  {
    // split into upstream noncoding exon and prc
    int begNpc=-1,endNpc=-1,begPrc=-1,endPrc=-1,bframePrc=-1,eframePrc=-1;
    if( (ip.first > (left.getRght()+1) && ip.first <= rght.getLeft()-1) ) {
      assert( ! (ip.second > left.getRght() && ip.second <= rght.getLeft()-1)) ;
      begNpc = left.getRght()+1>=4?left.getRght()+1:4;
      endNpc = ip.first-1;
      begPrc = ip.first;
      endPrc = rght.getLeft()-1;
      bframePrc = 0;
      eframePrc = mod2frame((rght.getLeft()-ip.first) % 3);
    } else if( ip.second > (left.getRght()+1) && ip.second <= (rght.getLeft()-1) ) {
      begNpc = ip.second+1;
      endNpc = rght.getLeft()-1;
      begPrc = left.getRght()+1>=4?left.getRght()+1:4;
      endPrc = ip.second;
      bframePrc = targetCoords.getFrame(begPrc,left,rght);
      if( bframePrc == -1 ) {
	cout<<"1turns out to be in npc portion of source transcript: "<<left<<" "<<rght<<endl;
	return false;
      }
      eframePrc = 2;
    }
    if( endPrc > begPrc ) {
      bmPrc.train(svec,begPrc,endPrc,bframePrc,eframePrc);
      const double prob = bmPrc.getLastProb();
      prob_sum[mnPrc] += prob;
      ++prob_cnt[mnPrc];
      cout<<"prc1 tprob: "<<ModelType::mname2str(mnPrc)<<" "<<prob<<endl;
    }
    if( endNpc > begNpc ) {
      bmNpc.train(svec,begNpc,endNpc,bframeNpc,eframeNpc);
      const double prob = bmNpc.getLastProb();
      prob_sum[mnNpc] += prob;
      ++prob_cnt[mnNpc];
      cout<<"npc1 tprob: "<<ModelType::mname2str(mnNpc)<<" "<<prob<<endl;
    }
    // split into upstream prc exon and downstream npc
  } else {
    bool isPrc = false;
    if( isExon && left.getRght()+1 >= (ip.first+2) && left.getLeft()-1 <= ip.second ) {
      isPrc=true;
    } else if(isExon) {
      isPrc=false;
    } else {
      //intron
    }
    const int beg = left.getRght()+1>=4?left.getRght()+1:4;
    const int end = rght.getLeft()-1;
    if( beg > end ) {
      cout<<"Overlapping intervals, skipping: "<<left<<" "<<rght<<endl;
      return false;
    }
    if(isPrc) {
      const int bfr = targetCoords.getFrame(beg,left,rght);
      if( bfr == -1 ) {
	cout<<"turns out to be in npc portion of source transcript: "<<left<<" "<<rght<<endl;
	return false;
      }
      const int tb = 3-bfr==3 ? 0 : 3-bfr;
      const int len = (end+1)-(beg-tb);
      if( len > 0 ) {
	const int efr = mod2frame( (len% 3) );
	bmPrc.train(svec,beg,end,bfr,efr);
	const double prob = bmPrc.getLastProb();
	prob_sum[mnPrc] += prob;
	++prob_cnt[mnPrc];
	cout<<"prc2 tprob: "<<ModelType::mname2str(mnPrc)<<" "<<prob<<endl;
      }
    } else {
      const int bfr = -1;
      const int efr = -1;
      bmNpc.train(svec,beg,end,bfr,efr);
      const double prob = bmNpc.getLastProb();
      prob_sum[mnNpc] += prob;
      ++prob_cnt[mnNpc];
      cout<<"npc2 tprob: "<<ModelType::mname2str(mnNpc)<<" "<<prob<<endl;
    }
  }
  return false;
}

bool 
isConsCodonCheck(unsigned pos, int frame, const vector<string*>& svec) {
  // only preclude the use of stop codons in target sequence
  int off =  -1;
  bool isConsCodon = true;
  if( frame==0 ) {
    off=1;
  } else if( frame == 1 ) {
    off=2;
  } else if( frame == 2 ) {
    off=3;
  }
  const int codonSize = 3;
  if( pos-off >= 0 && ((pos-off)+codonSize < svec[0]->size()) ) {
    assert( (pos-off) >= 0 && ((pos-off)+codonSize < svec[0]->size()));
    string codon = svec[0]->substr(pos-off,codonSize);
    makelower(codon);
    const char targAA = codon[0]!='-'&&codon[1]!='-'&&codon[2]!='-' ? codon2aa(codon) : 'B';
    for(unsigned k = 1; k < svec.size(); ++k) {
      if(svec[k]) {
	const string& tstr = *svec[k];
	assert( (pos-off) >= 0 && ((pos-off)+codonSize < svec[k]->size()));
	string icodon = tstr.substr(pos-off,codonSize);
	makelower(icodon);
	const char infAA = icodon[0]!='-'&&icodon[1]!='-'&&icodon[2]!='-' ? codon2aa(icodon) : 'B';
	if( targAA != infAA ) {
	  isConsCodon = false;
	  break;
	}
      }
    }
  }
  return isConsCodon;
}

spair_t
ModelManager::compProb1(unsigned pos, int frame, const vector<string*>& svec, const string& column, cname_t cn, 
			const string& str, pred_t predType, bool isInner, int start_codon, bool verbose) {
  const bool isPrcExon = ModelType::isPrcExon(cn);
  bool isConsCodon = true;
  if(isPrcExon) {
    // only preclude the use of stop codons in target sequence
    int off =  -1;
    // why frame=0,1 case here? 
    if( frame==0 ) {
      off=1;
    } else if( frame == 1 ) {
      off=2;
    } else if( frame == 2 ) {
      off=3;
    }
    const int codonSize = 3;
    if( pos-off >= 0 && ((pos-off)+codonSize < svec[0]->size()) ) {
      assert( (pos-off) >= 0 && ((pos-off)+codonSize < svec[0]->size()));
      string codon = svec[0]->substr(pos-off,codonSize);
      makelower(codon);
      if(frame == 2 ) {
        if(codon=="taa" || codon=="tga"||codon=="tag") {
          return spair_t(false,0);
        }
      }
      isConsCodon = isConsCodonCheck(pos,frame,svec);
    }
  }
  // use conserved model only
  const int prev = 3;
  bool isConsCol = isCons(pos,svec,prev);
  int clevel = 0;
  if( isPrcExon && isConsCodon && !isConsCol ) {
    clevel = 3; // tolerate synonymous substitutions
  } else if( isPrcExon || ModelType::isSplice(cn) ) {
    // otherwise more strict for codons and splice sites
    clevel = 0;
  } else if( isConsCol ) {
    // otherwise if conserved column, use slow mutation model
    clevel = 0;
  } else {
    // otherwise if not conserved use fast mutation model
    clevel = 3;
  }
  /** try to ignore the scoring deal **/
  const bool val = ModelType::isSplice(cn) ? true : clevel==0;
  int tval = 0;
  if( _sepNonCons && ModelType::isIntron(cn) ) {
    //tval = ModelType::isSplice(cn) ? 0 : clevel;
    //only distinguish intron conservation
    tval = clevel;
  }
  ModelType::mname_t con_mn = ModelType::cname2mname(cn,tval);
  MultiModel* tmcm = _mm[con_mn];

  ModelType::mname_t mn = ModelType::cname2mname(cn,clevel);
  PhyloModel* pm = _pm[mn];
  const int sidx =  val ? 0 : 1;
  double prate = _rm->getRateHmm(cn)->getState(sidx).getRate();
#if 0
  /// hack to test potential of this approach
  if( ModelType::isSplice(cn) && !isInner ) {
    ModelType::cname_t tmp = ModelType::isDon(cn) ? ModelType::eCintronUp : ModelType::eAintronUp;
    prate = _rm->getRateHmm(tmp)->getState(0).getRate();
  }
  if( pos == 104 && cn == ModelType::eCdon ) {
    GDEBUG=true;
  } else {
    GDEBUG=false;
  }
  if( XDEBUGON ) {
    int stop = 0;
  }
#endif
  double cprob = predType==eLearnedFixedRate ? pm->getProb(svec,pos,frame,column,tmcm,prate) : pm->getProb(svec,pos,frame,column,tmcm);
  ModelType::mname_t con_fn = ModelType::getFalseName(mn,val);
  MultiModel* fmcm = _mm[con_fn];
  ModelType::mname_t fn = ModelType::getFalseName(mn,val);
  PhyloModel* fm = _pm[fn];
  const double frate = 1.0;
  double fprob = predType==eLearnedFixedRate ? fm->getProb(svec,pos,frame,column,fmcm,frate) : fm->getProb(svec,pos,frame,column,fmcm);
  bool status = true;
  double nsum = 0;
  if( fprob == 0 || cprob == 0 ) {
    //cout<<"Warning something wrong : fprob="<<fprob<<" cprob="<<cprob<<endl;
    status = false;
  } else {
    nsum = log2(cprob/fprob);
  }
#if 0
  if( XDEBUGON ) {
    cout<<"pos: "<<pos<<" fr: "<<frame<<" "<<ModelType::cname2str(cn)<<" "<<cprob<<" "<<fprob<<" "<<prate<<" "<<frate<<" status: "<<(status?1:0)<<" res: "<<nsum<<endl;
  }
  //cout<<"ctype: "<<ModelType::cname2str(cn)<<" score "<<nsum<<endl;
#endif
  return spair_t(status,nsum);
}

tp_t
ModelManager::getCname(const list<SeqLabel>& lst, int pos, bool isPrc) {
  tp_t ret(ModelType::eCnameEnd,-1);
  list<SeqLabel>::const_iterator ti=lst.begin();
  const list<SeqLabel>::const_iterator ts=lst.end();
  const SeqLabel* lptr = NULL;
  for(; ti != ts; ++ti) {
    const SeqLabel& sl = *ti;
    if( sl.isIn(pos) && !sl.isEnd() ) {
      ret.first = ModelType::sl2cnamePos(sl);
      ret.second = pos-sl.getLeft();
    } else if( lptr ) {
      if( pos >= lptr->getRght()+1 && pos < sl.getLeft() || (pos == sl.getLeft() && sl.isEnd())  
	  ) {
	ret.first = ModelType::sl2cnameInv(*lptr,sl,isPrc);
	ret.second = -1;
	break;
      }
    }
    lptr = &sl;
  }
  assert(ret.first != ModelType::eCnameEnd);
  return ret;
}

#if 1
void
ModelManager::compare(const vector<string*>& svec, int start, int stop, const vector<list<SeqLabel>*>& cvec, pred_t predType) {
  vector<vector<int> > ftrack(3);
  for(unsigned d=0;d<ftrack.size(); ++d) {
    ftrack[d].resize(cvec.size(),d);
  }
  vector<vector<spair_t> > psum(3);
  for(unsigned d=0;d<psum.size(); ++d) {
    psum[d].resize(cvec.size(),spair_t(true,0));
  }
  int left = cvec[0]->front().getLeft()+1;
  int rght = cvec[0]->back().getRght()-1;
  bool lastPrc = (left >= start && left <=stop);
  for(signed i = left; i <= rght; ++i) {
    if( i < 4 ) continue; 
    const string& str = *svec[0];
    if( str[i] == '-' ) {
      continue;
    }
    bool isPrc = isPrcCheck(i,start,stop);
    //bool isPrc = (i >= start && i<= stop);
    const unsigned num_org = svec.size();
    string column(num_org,'-');
    fillColumn(column,svec,i);
    cout<<"pos: "<<i<<" ";
    cout<<column<<endl;
    for(unsigned x = 0; x < psum.size(); ++x) {
      for(unsigned k = 0; k < cvec.size(); ++k) {
	const list<SeqLabel>& lst = *cvec[k];
	const tp_t tp = getCname(lst,i,isPrc);
	const cname_t cn = tp.first;
	const int rframe = tp.second;
	if(  x==0 && !lastPrc && isPrc && ModelType::isPrcExon(cn)) {
	  for(unsigned tk = 0; tk < ftrack[0].size(); ++tk) {
	    ftrack[0][tk]=0;
	    ftrack[1][tk]=ftrack[2][tk]=-1;
	  }
	}
	int& frame = ftrack[x][k];
	//cout<<ftrack[0][k]<<" "<<ftrack[1][k]<<" "<<ftrack[2][k]<<endl;
	//assert(ftrack[0][k] != ftrack[1][k] && ftrack[0][k] != ftrack[2][k]);
	if( psum[x][k].first ) {
	  // if non-coding you don't need to compute for different frame
	  if( !ModelType::isPrcExon(cn) && x > 0 && psum[0][k].first ) {
	    psum[x][k] = psum[0][k];
	  } else {
	    const int phase = ModelType::isPrcExon(cn) ? frame : rframe;
	    if( ModelType::isPrcExon(cn) && phase == -1) {
	      continue; //assert(phase != -1);
	    }
	    const bool pleaseSetMe=false;
	    const spair_t score =compProb1(i,phase,svec,column,cn,str,predType,pleaseSetMe,start);
	    if( score.first ) {
	      psum[x][k].second += score.second;
	      cout<<"["<<ModelType::cname2str(cn)<<" "<<k<<" f:"<<x<<" p:"<<phase<<" s:"<<score<<" "<<psum[x][k].second<<"]";
	    } else {
	      cout<<k<<" frame: "<<x<<" "<<rframe<<" crash and burn\n";
	      psum[x][k].first = false;
	    }
	  }
	}
	if( !ModelType::isIntron(cn) ) {
	  frame++;
	  frame = frame > 2 ? 0 : frame;
	}
      }
    }
    lastPrc = isPrc;
    cout<<endl;
  }
  for(unsigned x = 0; x < psum.size(); ++x) {
    for(unsigned k = 0; k < ModelType::eInvCnameEnd; ++k) {
      cout<<"frame: "<<x<<" k: "<<k<<" sum: "<<psum[x][k]<<endl;
    }
  }
}
#endif

vector<spair_t>
ModelManager::scoreInterval1a(const SeqLabel& left, const SeqLabel& rght, const vector<string*>& svec, int start, int stop, 
			      const SeqLabel* lsl, const SeqLabel* rsl, pred_t predType, bool verbose) {
  bool debug=false;
  //if( left.getLeft()==10 && left.getLabel()==SeqLabel::eFma1 && rght.getLeft()==79 && rght.getLabel()==SeqLabel::eFma2) {
    //debug=true;
  //}
  const int spos = lsl ? lsl->getRght()+1 : left.getRght()+1;
  const int rpos = rsl ? rsl->getRght()+1 : rght.getLeft();
  vector<int> ftrack(3);
  //getPhases(ftrack,spos,start);
#if 1
  if( lsl ) {
    getPhases(ftrack,spos,start);
  } else {
    getPhases(ftrack,left,start);
  }
#endif
  const bool iex = SeqLabel::isExon(left.getLabel(),rght.getLabel());
  bool lastPrc = false;
  vector<spair_t> psum(3,spair_t(true,0));

  if( rsl && !lsl && (left.getLabel()==SeqLabel::eFmd2 || left.getLabel()==SeqLabel::eAfmd) ) {
	  	//check splice window for stop codons
    for(signed i = rght.getLeft(), relPos = 0; i <= rght.getRght(); ++i, ++relPos) {
      const bool isPrc = isPrcCheck(i,start,stop);
  		const cname_t cn = ModelType::sl2cnameInv(left,rght,isPrc);
      //const bool isPrc = (i >= start && i<= stop);
      if( isPrc && ModelType::isExon(cn) ) {
    	   vector<int> spl_track(3);
         getPhases(spl_track,i,start);
         for(unsigned x = 0; x < 3; ++x) {
            int& frame = spl_track[x];
            if( frame == 2 ) {
              int off =  3;
              const int codonSize = 3;
              if( i-off >= 0 && ((i-off)+codonSize < (signed)svec[0]->size()) ) {
                assert( (i-off) >= 0 && ((i-off)+codonSize < (signed)svec[0]->size()));
                string codon = svec[0]->substr(i-off,codonSize);
                makelower(codon);
                if(codon=="taa" || codon=="tga"||codon=="tag") {
                   psum[x].first = false;
                }
              }
           }
        }
     }
    }
  } 

  for(signed i = spos; i < rpos; ++i) {
    if( i < 4 ) continue;
    const string& str = *svec[0];
    if( str[i] == '-' ) {
      continue;
    }
    const bool isPrc = isPrcCheck(i,start,stop);
    //bool isPrc = (i >= start && i<= stop);
    const cname_t cn = ModelType::sl2cnameInv(left,rght,isPrc);
    if( cn == ModelType::eCnameEnd ) {
      return vector<spair_t>(3,spair_t(false,0));
    }
    if(!isPrc || !iex) {
      ftrack[0]=ftrack[1]=ftrack[2] = -1;
    }
    // this means we've switch into prc
    // and thie 1st time, and we know only frame=0 is valid
    else if(  ftrack[0] == -1 && isPrc && iex ) {
      ftrack[0]=0;
      ftrack[1]=ftrack[2]=-1;
    }
    const unsigned num_org = svec.size();
    string column(num_org,'-');
    fillColumn(column,svec,i);
    if(debug)cout<<"column "<<column<<" "<<i<<endl;
    spair_t pscore(false,0);
    for(unsigned x = 0; x < 3; ++x) {
      int& frame = ftrack[x];
      if( frame == -1 && iex && isPrc ) {
	psum[x] = spair_t(false,0);
	continue;
      }
      // if non-coding you don't need to compute for different frame
      if( frame == -1 && x > 0 && psum[x].first ) {
	assert(pscore.first);
	psum[x].second += pscore.second;
      } else if( psum[x].first || (frame == -1 && x==0)) {
	//cout<<"lframe: "<<x<<endl;
	const spair_t score =compProb1(i,frame,svec,column,cn,str,predType,false,start,verbose);
	pscore = score;
	//if(GDEBUG) {
	//  cout<<column<<" thispos: "<<i<<" "<<score<<" "<<frame<<endl;
	//}
	if( score.first && psum[x].first ) {
	  psum[x].second += score.second;
	} else {
	  psum[x] = spair_t(false,0);
	  //if(verbose) {
	  //cout<<"stop codon failed on frame: "<<x<<" pos: "<<i<<" phase: "<<frame<<endl;
	  //}
	}
      }
      //if(verbose && x==0 ) {
	//cout<<"pos: "<<i<<" phase: "<<frame<<" "<<pscore<<" "<<psum[x]<<endl;
      //}
      if( isPrc && iex ) {
	frame++;
	frame = frame > 2 ? 0 : frame;
      }
    }
    lastPrc = isPrc;
    if( !psum[0].first && !psum[1].first && !psum[2].first) {
      break;
    }
    //assert(psum[1] != 2);
  }
  if( !rsl && !rght.isEnd() && rght.getLabel() != SeqLabel::eBeg ) {
    vector<int> spl_track(3);
    bool isInner=SeqLabel::isInner(rght.getLabel());
    cname_t splMod = ModelType::sl2cnamePos(rght);
    assert(splMod != ModelType::eCnameEnd);
    for(signed i = rght.getLeft(), relPos = 0; i <= rght.getRght(); ++i, ++relPos) {
      const string& str = *svec[0];
      if( i >= (signed)str.size() ) {
	break;
      }
      if( i < 2 ) continue; // oogilly, come back and fix - assumes 1st order models
#if 0
      if( i== 79 && left.getLeft()==10 && left.getLabel()==SeqLabel::eFma1 && rght.getLeft()==79 && rght.getLabel()==SeqLabel::eFma2) {
	int stop = 0;
      }
#endif

      // assumes no overlapping donor windows
      // if such occase occurs may mistakenly read a non-existent stop codon
      // the fix requires looking for stop codons on the next pass
      // i.e. if this is efmd1 then you can't check it until efAfmd1.
      bool isPossibleExon = true;
      if( i < left.getExonBnd() && SeqLabel::isAcc(left.getLabel()) && !SeqLabel::isAccIr(left.getLabel()) ) {
	 		isPossibleExon =false;
      }
	   if(SPECIAL_DEBUG && SeqLabel::isAcc(rght.getLabel()) && i < rght.getExonBnd() ) {
			isPossibleExon=false;
      }
      if( rght.getLabel()==SeqLabel::eFmd2 || rght.getLabel()==SeqLabel::eAfmd)  {
			isPossibleExon=false;
      }
      const bool isPrc = isPrcCheck(i,start,stop);
      //const bool isPrc = (i >= start && i<= stop);
      if( isInner && isPrc && isPossibleExon ) {
	const cname_t cn = ModelType::sl2cnameInv(left,rght,isPrc);
	if( !ModelType::isExon(cn) ) {
	  assert(SPECIAL_DEBUG);
	}
	getPhases(spl_track,i,start);
	for(unsigned x = 0; x < 3; ++x) {
	  int& frame = spl_track[x];
	  if( frame == 2 ) {
	    int off =  3;
	    const int codonSize = 3;
	    if( i-off >= 0 && ((i-off)+codonSize < (signed)svec[0]->size()) ) {
	      assert( (i-off) >= 0 && ((i-off)+codonSize < (signed)svec[0]->size()));
	      string codon = svec[0]->substr(i-off,codonSize);
	      makelower(codon);
	      if(codon=="taa" || codon=="tga"||codon=="tag") {
		psum[x].first = false;
	      }
	    }
	  }
	}
      }
      const unsigned num_org = svec.size();
      string column(num_org,'-');
      fillColumn(column,svec,i);
      const spair_t score = compProb1(i,relPos,svec,column,splMod,str,predType,isInner,start);
      for(unsigned x = 0; x < 3; ++x) {
	if( score.first && psum[x].first ) {
	  psum[x].second += score.second;
	} else {
	  psum[x].first = false;
	}
      }
    }
  }
  return psum;
}

vector<spair_t>
ModelManager::scoreInterval1b(const SeqLabel& left, const SeqLabel& rght, const vector<string*>& svec, int start, int stop,
			      const SeqLabel* lsl, const SeqLabel* rsl, bool verbose) {
  if(svec.size()==1) {
    //return scoreInterval1a(left,rght,svec,start,stop,NULL,NULL,verbose);
    assert(0);
  }
  const int spos = lsl ? lsl->getRght()+1 : left.getRght()+1;
  const int rpos = rsl ? rsl->getRght()+1 : rght.getRght()+1;
  vector<int> ftrack(3);
  if( lsl ) {
    getPhases(ftrack,spos,start);
  } else {
    getPhases(ftrack,left,start);
  }
  //const bool iex = SeqLabel::isExon(left.getLabel(),rght.getLabel());
  //bool lastPrc = false;
  vector<spair_t> psum(3,spair_t(true,0));
  vector<Trellis*> cTrellis;
  vector<htype_t> save_ht, last_ht;
  bool isInit=true;
  for(signed i = spos; i < rpos; ++i) {
    if( i < 4 ) continue; 
    if( i >= (signed)svec[0]->size() ) break;
    if( !psum[0].first && !psum[1].first && !psum[2].first ) break;
    const bool isPrc = isPrcCheck(i,start,stop);
    //bool isPrc = (i >= start && i<= stop);
    cname_t cn = ModelType::eCnameEnd;
    if( rght.isIn(i) ) {
      cn = ModelType::sl2cnamePos(rght);
    } else {
      cn = ModelType::sl2cnameInv(left,rght,isPrc);
    }
    if( cn == ModelType::eCnameEnd ) {
      return vector<spair_t>(3,spair_t(false,0));
    }
    scorePosHelp2(false,i,svec,cn,last_ht,isInit,ftrack,psum,cTrellis);
    save_ht = last_ht;
    isInit=false;
  }
  //kinda like to know if null for now
  if( rpos < 4 || (rpos-spos)< 1 ) {
    psum[0]=psum[1]=psum[2]=spair_t(false,0);
  }
  if( rpos > 4 && rpos-spos>0 ) {
    assert(!cTrellis.empty());
  }
  if(!cTrellis.empty()) {
    for(unsigned x = 0; x < 3; ++x) {
      const int cphase = ftrack[x]==-1?0:ftrack[x];
      unsigned idx = cTrellis.size()==1 ? 0 : x;
      htype_t sht = save_ht[idx];
      if( sht != eEmptyState && psum[x].first ) {
	const int pphase = Trellis::getPrevTime(sht,cphase);
	const spair_t sp = cTrellis[idx]->getBestScore(sht,pphase);
	if( sp.first ) {
	  psum[x].first = true;
	  psum[x].second += sp.second;
	}
      }
    }
    for(unsigned x = 0; x < cTrellis.size(); ++x) {
      if(cTrellis[x])
	delete cTrellis[x];
    }
  }
  return psum;
}
