#include "PhyloModel.h"
#include "utils.h"
#include "MarkovChain.h"
#include "TamuraNei.h"
#include "stype.h"
#include "base.h"
#include "cidx.h"
#include <iostream>
#include <cmath>
#include <cctype>
#include <sstream>

using bioinfo::idx2c;
using std::string;
using std::vector;
using std::list;
using std::cout;
using std::cerr;
using std::endl;

static const unsigned cBASE = 5;
static const double small = 0.0001;

extern bool YDEBUGON;
extern bool GDEBUG;
extern bool DEBUGON;

PhyloModel::PhyloModel(const Phylogeny& phy, double mrate, int npos)
  : _phy(NULL), _tnTvRatio(2.0), _mutRate(mrate) {
    _phy = phy.copy();
}

PhyloModel::~PhyloModel() {
  delete _phy;
}


static void
makeKey(const vector<string*>& svec, int pos, int time, int order, double mutRate, string& keyStr) {
  if( ((pos-1)-order ) >= 0 ) {
    signed width = order+1;
    for(unsigned x = 0, istart=0; x < svec.size(); ++x, istart+=width) {
      if( svec[x] ) {
	const string& gstr = *svec[x];
	signed gidx = (pos-1), ccnt = 0;
	for(signed idx = istart; gidx >= 0 && ccnt <= width && idx < (signed)istart+width; ) {
	  assert(gidx >= 0 && gidx < (signed)gstr.size());
	  const char tchr = gstr[gidx];
	  if( !std::islower(tchr) && (tchr != '-' || gidx==(pos-1)) ) {
	    assert(idx >= 0 && idx < (signed)keyStr.size());
	    keyStr[idx]=tchr;
	    ++ccnt;
	    --gidx;
	    ++idx;
	  } else {
	    --gidx;
	  }
	}
      }
    }      
    if( time < 0 )
      time = 0;
    unsigned sidx = (svec.size()*(order+1));
    keyStr[sidx] = getDigitPos0(time);
    int off=0;
    if(time >= 10) {
      ++off;
      assert(time < 100); // one char per digit (assume max of 2 digits for now);
      keyStr[sidx+off] = getDigitPos1(time);
    }
    assert(mutRate >0 && mutRate <10); //assume a 2 digit number
    int ratei = (int)(mutRate*10.0);
    keyStr[sidx+off+1] = getDigitPos0(ratei);
    if(ratei >= 10) {
      assert(ratei<100);
      keyStr[sidx+off+2] = getDigitPos1(ratei);
    }
  }
}

const SubModel*
PhyloModel::buildModel(tnode* node) const {
  if( !node->_subHash ) {
    node->_subHash = new Phylogeny::shash_t;
  }
#if 0
  if(GDEBUG) {
    cout<<"NODE: "<<node->_id<<" blen: "<<node->_blen;
    list<int>::const_iterator iter = node->_dlst.begin();
    const list<int>::const_iterator stop = node->_dlst.end();
    for(; iter != stop; ++iter) {
      cout<<" "<<*iter;
    }
    cout<<endl;   
  }
#endif
  return _tgMultiModel->getMc(*_tgSvec,_tgPos,_tgTime,*node->_subHash,node->_dlst,_tnTvRatio,_mutRate);
}

PhyloModel::ppair2_t
PhyloModel::calcProbHelp2(tnode* node, const string& str, char anc, double rate, bool debug) {
  bool status = true;
  double prob= 0;
  if(node->_left && node->_rght) {
    double sum=0;
    for(unsigned i = 0; i < 4; ++i) {
      const char lb = bioinfo::idx2c(i);
      const SubModel* sm = buildModel(node->_left);
      double lp = sm ? sm->getProb(lb,anc,node->_left->_blen*rate): 0;
      if( lp==0 && sm ) {
	continue;
      }
      const ppair2_t ln = calcProbHelp2(node->_left,str,lb,rate,debug);
      if( ln.first == 0 && ln.second ) {
	continue;
      }
#if 0
      if(GDEBUG) {
	cout<<"left prob: "<<anc<<" to "<<lb<<" blen: "<<node->_left->_blen<<" id: "<<node->_id<<" "<<lp<<" res: ";
	cout<<ln.first<<" "<<ln.second<<endl;
      }
#endif
      for(unsigned j = 0; j < 4; ++j) {
	const char rb = bioinfo::idx2c(j);
	const SubModel* sm = buildModel(node->_rght);
	double rp = sm ? sm->getProb(rb,anc,node->_rght->_blen*rate) : 0;
	if( rp == 0 && sm ) {
	  continue;
	}
	const ppair2_t rn = calcProbHelp2(node->_rght,str,rb,rate,debug);
#if 0
	if(GDEBUG) {
	  cout<<"rght prob: "<<anc<<" to "<<rb<<" blen: "<<node->_rght->_blen<<" id: "<<node->_id<<" "<<rp<<" res: ";
	  cout<<rn.first<<" "<<rn.second<<endl;
	}
#endif
	if( ln.second && rn.second ) {
	  sum += (lp*ln.first*rp*rn.first);
	}
	else if( ln.second ) {
	  sum += (lp*ln.first);
	  break;
	}
	else if( rn.second ) {
	  sum += (rp*rn.first);
	}
	if( !ln.second && !rn.second) {
	  status = false;
	  break;
	}
      }
      if( !ln.second ) {
	break;
      }
    }
    prob = sum;
  } else {
    // leaf
    assert(node->_idx != -1);
    const char leaf = tolower(str[node->_idx]);
    if( leaf == '-' ) { 
      // prob = 0;
      status = false;
    }
    else if( leaf == anc ) {
      prob = 1;
    } else {
      //prob = 0; // can save time 
      //initialized to 
    }
  }
#if 0
  if(GDEBUG) {
    cout<<"prob: "<<prob<<" status: "<<status<<endl;
  }
#endif
  return ppair2_t(prob,status);
}

double
PhyloModel::getProb(const vector<string*>& svec, int pos, int time, const string& column, MultiModel* multiModel, double mutRate) {
  if(mutRate != -1) {
    _mutRate = mutRate;
  }
  // assume time=-1 implies just a single model (i.e. index=0);
  if( time == -1 )
    time = 0;
#if 0
  if( pos == 274 && YDEBUGON ) {
    int stop = 0;
    GDEBUG=true;
  } else {
    GDEBUG=false;
  }
#endif
  const int order = multiModel->getOrder();
  assert(time < 100); // one char per digit (assume max of 2 digits for now);
  unsigned nSize1 = time >= 0 && time <= 9 ? 1 : 2;
  unsigned nSize2 = _mutRate < 1 ? 1 : 2;
  unsigned strSize = nSize1+nSize2+(svec.size()*(order+1));
  string keyStr(strSize,'-');
  makeKey(svec,pos,time,order,_mutRate,keyStr);
  double score = _cache[keyStr]._dat;
  if(GDEBUG) {
    cout<<"ph score: "<<score<<" key: "<<keyStr<<endl;
  }
  if( score == -1 || GDEBUG) {
    //if( score == -1 ) {
    const bool debug=false;
    _tgSvec = &svec;
    _tgPos = pos;
    _tgTime = time;
    _tgMultiModel = multiModel;
    double prob = calcProb(svec,pos,time,column,multiModel,debug);
    score = prob;
    _cache[keyStr] = score;
  }
  if(GDEBUG) {
    cout<<"score: "<<score<<" "<<pos<<" "<<time<<" "<<column<<endl;
    exit(0);
  }
  return score;
}

double
PhyloModel::calcProb(const vector<string*>& svec, int pos, int time, const string& column, MultiModel* multiModel, bool debug) {
  double sum = 0;
  const MarkovChain* mc = multiModel->getMc(svec,pos,time);
  if(!mc) {
    return sum;
  }
  if( svec.size()== 1 ) {
    assert(column.size()==1);
    string cstr(1,column[0]);
    const double val = mc->getVal(cstr,1);
    sum=val;
    return sum;
  }
  const int treeIdx = 0;
  if( mc ) {
    for(unsigned i = 0; i < 4; ++i) {
      const char ch = bioinfo::idx2c(i);
      string cstr(1,ch);
      const double anc_pch = mc->getVal(cstr,1);
      if( anc_pch > 0 ) {
	Phylogeny::tnode* rt = _phy->getRoot(treeIdx);
#if 0
	if(GDEBUG) {
	  int stop = 0;
	}
#endif
	ppair2_t pp = calcProbHelp(rt,column,ch,_mutRate,debug);
	assert(pp.second);
	const double val = anc_pch*pp.first;
	if(GDEBUG) {
	  cout<<"check: "<<val<<" "<<anc_pch<<" "<<pp.first<<endl;
	}
	sum += val;
      }
    }
    if( sum > 1 ) {
      cout<<"Warning some kind of overflow? "<<"["<<column<<"]"<<" score: "<<sum<<endl;
      sum = 1;
    }
    if( sum < 0 ) {
      cout<<"Warning some kind of overflow? "<<"["<<column<<"]"<<" score: "<<sum<<endl;
      sum = 0;
    }
  }
  if( sum == 0 ) {
    // calcProb(svec,pos,time,column,multiModel,debug);
  }
  return sum;
}

PhyloModel::ppair2_t
PhyloModel::calcProbHelp(tnode* node, const string& str, char anc, double rate, bool debug) {
  bool status = true;
  double lsum= 0, rsum = 0;
  bool ignoreLeft = false, ignoreRght = false;
  if(node->_left) {
    for(unsigned i = 0; i < 4; ++i) {
      const char lb = bioinfo::idx2c(i);
      const SubModel* sm = buildModel(node->_left);
      double lp = sm ? sm->getProb(lb,anc,(node->_left->_blen*rate)): 0;
      if( lp==0 && sm ) {
	continue;
      }
      const ppair2_t lprob = calcProbHelp(node->_left,str,lb,rate,debug);
      if( !lprob.second ) {
	ignoreLeft = true;
	break;
      }
      lsum += (lp*lprob.first);
#if 1
      if(GDEBUG) {
	cout<<"left prob: "<<anc<<" to "<<lb<<" blen: "<<node->_left->_blen<<" id: "<<node->_id<<" "<<lp<<" res: ";
	cout<<lprob.first<<" "<<lprob.second<<endl;
      }
#endif
    }
  } 
  if(node->_rght) {
    for(unsigned j = 0; j < 4; ++j) {
      const char rb = bioinfo::idx2c(j);
      const SubModel* sm = buildModel(node->_rght);
      double rp = sm ? sm->getProb(rb,anc,(node->_rght->_blen*rate)) : 0;
      if( rp == 0 && sm ) {
	continue;
      }
      const ppair2_t rprob = calcProbHelp(node->_rght,str,rb,rate,debug);
#if 1
      if(GDEBUG) {
	cout<<"rght prob: "<<anc<<" to "<<rb<<" blen: "<<node->_rght->_blen<<" id: "<<node->_id<<" "<<rp<<" res: ";
	cout<<rprob.first<<" "<<rprob.second<<endl;
      }
#endif
      if( !rprob.second) {
	ignoreRght = true;
	break;
      }
      rsum += (rp*rprob.first);
    }
  }
  double prob = 0;
  if( !node->_left && !node->_rght) {
    // leaf
    assert(node->_idx != -1);
    const char leaf = tolower(str[node->_idx]);
    if( leaf == '-' ) { 
      // prob = 0;
      status = false;
    }
    else if( leaf == anc ) {
      prob = 1;
    } else {
      //prob = 0; // can save time 
      //initialized to 
    }
  } else {
    if( ignoreLeft && !ignoreRght ) {
      prob = rsum;
    } else if( !ignoreLeft && ignoreRght) {
      prob = lsum;
    } else if( !ignoreLeft && !ignoreRght ) {
      prob = lsum*rsum;
    } else {
      prob = 0;
      status = false;
    }
  }
#if 1
  if(GDEBUG) {
    cout<<"prob: "<<prob<<" status: "<<status<<endl;
  }
#endif
  return ppair2_t(prob,status);
}
