#include "Phylogeny.h"
#include "MarkovChain.h"
#include "base.h"
#include "SubModel.h"
#include <fstream>
#include <stdexcept>
#include <sstream>
#include <iostream>
#include "cidx.h"

using std::endl;
using std::ostream;
using std::ifstream;
using std::ostringstream;
using std::istringstream;
using std::string;
using std::list;
using std::runtime_error;
using std::vector;
using std::cout;
using std::endl;

extern bool GDEBUG;

typedef std::pair<double,bool> ppair2_t;

static void
print_help(ostream& os, const string& buff, Phylogeny::tnode* node);

static string::size_type
getBranchLen(Phylogeny::tnode* parent, const char* buff) {
  string::size_type pos = 0;
  if( *buff != ';' ) {
    assert(*buff == ':');
    string tmp = buff;
    pos = tmp.find_first_of(",)");
    assert(pos != string::npos);
    string sub = tmp.substr(1,pos);
    double blen = -1;
    istringstream istrm(sub);
    istrm>>blen;
    assert(blen != -1);
    parent->_blen = blen;
  }
  return pos;
}

Phylogeny::tnode::~tnode() {
  if(_subHash) {
    shash_t::iterator iter = _subHash->begin();
    const shash_t::iterator stop = _subHash->end();
    for(; iter != stop; ++iter) {
      const SubModel* val = (*iter).second;
      delete val;
    }
    delete _subHash;
  }
}

Phylogeny::~Phylogeny() {
  del();
}

Phylogeny::tnode*
Phylogeny::buildTree(const OrgIdx& oidx, tnode* parent, const char* buff) const {
  assert(parent);
  if( *buff == ')' ) {
    string::size_type advance = getBranchLen(parent,buff+1);
    if( advance > 0 ) 
      return buildTree(oidx,parent->_parent,buff+advance+1);
    return parent;
  } else if( *buff == '(' ) {
    parent->_left = new tnode(parent,0);
    return buildTree(oidx,parent->_left,buff+1);
  } else if(*buff == ':') {
    assert(0);
    string tmp = buff;
    string::size_type pos = tmp.find_first_of(",)");
    assert(pos != string::npos);
    string sub = tmp.substr(1,pos);
    double blen = -1;
    istringstream istrm(sub);
    istrm>>blen;
    assert(blen != -1);
    parent->_blen = blen;
    return buildTree(oidx,parent,buff+pos);
  } else if(*buff == ',') {
    parent->_rght = new tnode(parent,0);
    return buildTree(oidx,parent->_rght,buff+1);
  } else if(buff) {
    string tmp = buff;
    string::size_type pos = tmp.find_first_of(",)");
    assert(pos != string::npos);
    string sub = tmp.substr(0,pos);
    string::size_type loc = sub.find(':');
    sub[loc]=' ';
    istringstream istrm(sub);
    string id;
    double blen = -1;
    istrm>>id>>blen;
    assert(blen != -1);
    const int idx = oidx.fname2idx(id);
    parent->_idx = idx;
    parent->_id = id;
    parent->_blen = blen;
    return buildTree(oidx,parent->_parent,buff+pos);
  } else {
    assert(0);
    return NULL;
  }
}

list<int>
Phylogeny::storeDescendants(tnode* parent) {
  if( !parent->_left && !parent->_rght ) {
    parent->_dlst.push_back(parent->_idx);
    return parent->_dlst;
  } else if( !parent->_left ) {
    list<int> rd = storeDescendants(parent->_rght);
    list<int>::iterator beg = parent->_dlst.begin();
    parent->_dlst.splice(beg,rd);
  } else if( !parent->_rght ) {
    list<int> ld = storeDescendants(parent->_left);
    list<int>::iterator beg = parent->_dlst.begin();
    parent->_dlst.splice(beg,ld);
  } else {
    list<int> ld = storeDescendants(parent->_left);
    list<int> rd = storeDescendants(parent->_rght);
    list<int>::iterator beg = parent->_dlst.begin();
    parent->_dlst.splice(beg,ld);
    beg = parent->_dlst.begin();
    parent->_dlst.splice(beg,rd);
  }
  return parent->_dlst;
}

static Phylogeny::tnode*
copyHelp(Phylogeny::tnode* parent, Phylogeny::tnode* orig) {
  if( !orig ) {
    return NULL;
  } else {
    Phylogeny::tnode* nn = new Phylogeny::tnode(parent,orig->_blen,orig->_id);
    nn->_idx = orig->_idx;
    nn->_dlst = orig->_dlst;
    nn->_left = copyHelp(nn,orig->_left);
    nn->_rght = copyHelp(nn,orig->_rght);
    return nn;
  }
}

Phylogeny*
Phylogeny::copy() const {
  Phylogeny* res = new Phylogeny(1);
  tnode* root = _root[0];
  tnode* nroot = copyHelp(NULL,root);
  res->_root[0] = nroot;
  return res;
}

void
delHelp(Phylogeny::tnode* orig) {
  if( orig ) {
    delHelp(orig->_left);
    delHelp(orig->_rght);
    delete orig;
  }
}

void
Phylogeny::del() {
  tnode* root = _root[0];
  delHelp(root);
}

void
Phylogeny::read(const OrgIdx& orgIdx, const string& file) {
  _root.resize(orgIdx.size(),NULL); //= new tnode(NULL,0);
  //_oidx = orgIdx;
  ifstream ifs(file.c_str());
  if( !ifs ) {
    cout<<"stopped on: "<<file<<endl;
    throw runtime_error("couldn't open file");
  }
  string org;
  const int bsize = 1024;
  char buff[bsize];
  ifs.getline(buff,bsize);
  _root[0] = new tnode(NULL,0);
  const char* cptr = buff;
  _root[0]->_id="root";
  _root[0] = buildTree(orgIdx,_root[0],cptr);
  storeDescendants(_root[0]);
}

static void
print_help(ostream& os, const string& buff, Phylogeny::tnode* node) {
  if(node) {
    os<<buff<<node->_id<<" blen: "<<node->_blen;
    list<int>::const_iterator iter = node->_dlst.begin();
    const list<int>::const_iterator stop = node->_dlst.end();
    for(; iter != stop; ++iter) {
      os<<" "<<*iter;
    }
    os<<endl;
    print_help(os,string(" ") + buff,node->_left);
    print_help(os,string(" ") + buff,node->_rght);
  }
}

void
Phylogeny::print(ostream& os) const {
  string buff;
  for(unsigned i = 0; i < _root.size(); ++i) {
    if( _root[i] ) {
      print_help(os,buff,_root[i]);
    }
  }
}
