#include "RateHmm.h"
#include "RateTrans.h"
#include "RateState.h"
#include "Trellis.h"
#include <list>
#include <sstream>

using std::istringstream;
using std::string;
using std::list;
using std::ostream;
using std::istream;
using std::endl;

RateHmm::RateHmm(const RateHmm& rh) : 
  _states(rh._states.size()), _numLevels(rh._numLevels), _type(rh._type) {
    for(unsigned i = 0; i < _states.size(); ++i) {
      _states[i] = new RateState(*(rh._states[i]));
      assert( (signed)i == _states[i]->getId() );
    }
  list<RateTrans*>::const_iterator iter = rh._trans.begin();
  const list<RateTrans*>::const_iterator stop = rh._trans.end();
  for(; iter != stop; ++iter) {
    const RateTrans* trans = *iter;
    const int toIdx = trans->getTo().getId();
    const int frIdx = trans->getFr().getId();
    RateState& to = *(_states[toIdx]);
    RateState& fr = *(_states[frIdx]);
    RateTrans* nt = new RateTrans(fr,to,trans->getProb(),0);
    to.getInLst().push_back(nt);
    fr.getOutLst().push_back(nt);
    _trans.push_back( nt );
  }
}

RateHmm::RateHmm(int num_states, cname_t cn) :
  _states(num_states), _type(cn) {
}

RateHmm::~RateHmm() {
  for(unsigned i = 0; i < _states.size(); ++i) {
    delete _states[i];
  }
  list<RateTrans*>::iterator iter = _trans.begin();
  const list<RateTrans*>::iterator stop = _trans.end();
  for(; iter != stop; ++iter) {
    delete *iter;
  }
}

HmmComponent*
RateHmm::find(const HmmComponent& hc) {
  HmmComponent* rval = NULL;
  const RateState* rs = dynamic_cast<const RateState*>(&hc);
  if( rs ) {
    for(unsigned i = 0; i < _states.size(); ++i) {
      if( *(_states[i]) == *rs ) {
	rval = _states[i];
	break;
      }
    }
  } else {
    const RateTrans* rt = dynamic_cast<const RateTrans*>(&hc);
    assert(rt);
    list<RateTrans*>::iterator iter = _trans.begin();
    const list<RateTrans*>::iterator stop = _trans.end();
    for(; iter != stop; ++iter) {
      if( **iter == *rt ) {
	rval = *iter;
	break;
      }
    }
  }
  assert(rval);
  return rval;
}

void
RateHmm::print(ostream& os) const {
  os<<"RateHmm print: "<<endl;
  write(os);
}

void 
RateHmm::write(ostream& os) const {
  os<<"States"<<endl;
  for(unsigned i = 0; i < _states.size(); ++i) {
    _states[i]->write(os);
  }
  os<<"Transitions"<<endl;
  list<RateTrans*>::const_iterator iter = _trans.begin();
  const list<RateTrans*>::const_iterator stop = _trans.end();
  for(; iter != stop; ++iter) {
    const RateTrans* trans = *iter;
    trans->write(os);
  }
}

void
RateHmm::read(istream& is) {
  char buff[512];
  list<RateState*> rslst;
  //list<RateTrans*> rtlst;
  while(is.getline(buff,512) ) {
    if( string(buff)=="States" ) {
    } else if( string(buff)=="Transitions" ) {
      break;
    } else {
      istringstream istrm(buff);
      RateState* rs = new RateState;
      rs->read(istrm);
      rslst.push_back(rs);
    }
  }
  _states.resize(rslst.size());
  list<RateState*>::iterator iter = rslst.begin();
  const list<RateState*>::iterator stop = rslst.end();
  for(unsigned i = 0; i < _states.size(); ++i, ++iter) {
    assert(iter != stop);
    RateState* rs = *iter;
    int idx = rs->getId();
    //cout<<"load "<<idx<<endl;
    _states[idx]=rs;
  }
  while(is.getline(buff,512) ) {
    if( !*buff ) {
      break;
    }
    //cout<<"read trans: "<<buff<<endl;
    istringstream istrm(buff);
    ipair_t ip;
    double prob=-1,cnt=-1;
    RateTrans::read(istrm,ip,prob,cnt);
    RateState* fr = _states[ip.first];
    RateState* to = _states[ip.second];
    assert(fr && to);
    //cout<<"here; "<<endl;
    //fr->write(cout);
    //to->write(cout);
    RateTrans* rt = new RateTrans(*fr,*to,prob,cnt);
    _trans.push_back(rt);
  }
  //_trans.resize(rtlst.size());
  list<RateTrans*>::iterator titer = _trans.begin();
  const list<RateTrans*>::iterator tstop = _trans.end();
  for(; titer != tstop; ++titer) {
    RateTrans* trans = *titer;
    RateState& fr = trans->getFr();
    fr.getOutLst().push_back(trans);
    RateState& to = trans->getTo();
    to.getInLst().push_back(trans);
  }
}

void
RateHmm::normalize(htype_t ht) {
  const unsigned times = Trellis::getTimes(ht);
  const unsigned num_levels = Trellis::getLevels(ht);
  for(unsigned time = 0; time < times; ++time ) { 
    double init_sum = 0;
    for(unsigned level = 0; level < num_levels; ++level) {
      const unsigned idx = (times*level)+time;
      const RateState* rs = _states[idx];
      init_sum += rs->getCnt1();
    }
    if( init_sum > 0 ) {
      for(unsigned level = 0; level < num_levels; ++level) {
	const unsigned idx = (times*level)+time;
	RateState* rs = _states[idx];
	double nProb = rs->getCnt1() / init_sum;
	rs->setProb(nProb);
	rs->reset();
      }
    }
  }
  for(unsigned si = 0; si < _states.size(); ++si) {
    RateState& state = *_states[si];
    list<RateTrans*>& outGoing = state.getOutLst();
    list<RateTrans*>::const_iterator iter = outGoing.begin();
    const list<RateTrans*>::const_iterator stop = outGoing.end();
    double trans_sum = 0;
    for(; iter != stop; ++iter) {
      const RateTrans* trans = *iter;
      trans_sum += trans->getCnt1();
    }
    if(trans_sum > 0 ) {
      list<RateTrans*>::iterator miter = outGoing.begin();
      const list<RateTrans*>::iterator mstop = outGoing.end();
      for(; miter != mstop; ++miter) {
	RateTrans* trans = *miter;
	double nProb = trans->getCnt1() / trans_sum;
	trans->setProb(nProb);
	trans->reset();
      }
    }
  }
}

#if 1
void
RateHmm::normalize1(htype_t ht) {
  const unsigned times = Trellis::getTimes(ht);
  const unsigned num_levels = Trellis::getLevels(ht);
  for(unsigned time = 0; time < times; ++time ) { 
    double min = 0;
    for(unsigned level = 0; level < num_levels; ++level) {
      const unsigned idx = (times*level)+time;
      const RateState* rs = _states[idx];
      if( rs->getCnt() < min ) {
	min = rs->getCnt();
      }
    }
    assert(min<=0);
    min*=-1;
    double init_sum = 0;
    for(unsigned level = 0; level < num_levels; ++level) {
      const unsigned idx = (times*level)+time;
      RateState* rs = _states[idx];
      const double cnt = rs->getCnt();
      const double oProb = rs->getInitProb();
      const double wght = (cnt+min);
      const double tot = oProb + wght;
      init_sum += tot;
      rs->setProb(tot);
      rs->reset();
    }
    for(unsigned level = 0; level < num_levels; ++level) {
      const unsigned idx = (times*level)+time;
      RateState* rs = _states[idx];
      const double nProb = rs->getInitProb() / init_sum;
      rs->setProb(nProb);
    }
  }
  for(unsigned si = 0; si < _states.size(); ++si) {
    RateState& state = *_states[si];
    list<RateTrans*>& outGoing = state.getOutLst();
    list<RateTrans*>::const_iterator iter = outGoing.begin();
    const list<RateTrans*>::const_iterator stop = outGoing.end();
    double min = 0;
    for(; iter != stop; ++iter) {
      const RateTrans* trans = *iter;
      const double cnt = trans->getCnt();
      if( cnt < min ) {
	min = cnt;
      }
    }
    iter = outGoing.begin();
    assert(min<=0);
    min*=-1;
    double trans_sum = 0;
    list<RateTrans*>::iterator miter = outGoing.begin();
    const list<RateTrans*>::iterator mstop = outGoing.end();
    for(; miter != mstop; ++miter) {
      RateTrans* trans = *miter;
      const double cnt = trans->getCnt();
      const double oProb = trans->getProb();
      const double wght = (cnt+min);
      const double tot = oProb + wght;
      trans_sum += tot;
      trans->setProb(tot);
      trans->reset();
    }
    if( trans_sum > 0 ) {
      miter = outGoing.begin();
      for(; miter != mstop; ++miter) {
	RateTrans* trans = *miter;
	const double nProb = trans->getProb() / trans_sum;
	trans->setProb(nProb);
      }
    }
  }
}
#endif

#if 0
void
RateHmm::normalize1(htype_t ht) {
  const unsigned times = Trellis::getTimes(ht);
  const unsigned num_levels = Trellis::getLevels(ht);
  for(unsigned time = 0; time < times; ++time ) { 
    double min = 0;
    for(unsigned level = 0; level < num_levels; ++level) {
      const unsigned idx = (times*level)+time;
      RateState* rs = _states[idx];
      double nProb = rs->getCnt();
      if( nProb > 0 ) {
	rs->setProb(nProb);
	rs->reset();
      }
    }
  }
  for(unsigned si = 0; si < _states.size(); ++si) {
    RateState& state = *_states[si];
    list<RateTrans*>& outGoing = state.getOutLst();
    list<RateTrans*>::const_iterator iter = outGoing.begin();
    const list<RateTrans*>::const_iterator stop = outGoing.end();
    double min = 0;
    for(; iter != stop; ++iter) {
      RateTrans* trans = *iter;
      double nProb = trans->getCnt();
      if( nProb > 0 ) {
	trans->setProb(nProb);
	trans->reset();
      }
    }
  }
}
#endif

void
RateHmm::normalize2(htype_t ht) {
  const unsigned times = Trellis::getTimes(ht);
  const unsigned num_levels = Trellis::getLevels(ht);
  for(unsigned time = 0; time < times; ++time ) { 
    double init_sum = 0;
    for(unsigned level = 0; level < num_levels; ++level) {
      const unsigned idx = (times*level)+time;
      RateState* rs = _states[idx];
      const double cnt = rs->getCnt();
      const double oProb = rs->getInitProb();
      const double tot = cnt==0 ? oProb : cnt;
      init_sum += tot;
      rs->setProb(tot);
      rs->reset();
    }
    if(init_sum > 0) {
      for(unsigned level = 0; level < num_levels; ++level) {
	const unsigned idx = (times*level)+time;
	RateState* rs = _states[idx];
	const double nProb = rs->getInitProb() / init_sum;
	rs->setProb(nProb);
      }
    }
  }
  for(unsigned si = 0; si < _states.size(); ++si) {
    RateState& state = *_states[si];
    list<RateTrans*>& outGoing = state.getOutLst();
    double trans_sum = 0;
    list<RateTrans*>::iterator miter = outGoing.begin();
    const list<RateTrans*>::iterator mstop = outGoing.end();
    for(; miter != mstop; ++miter) {
      RateTrans* trans = *miter;
      const double cnt = trans->getCnt();
      const double oProb = trans->getProb();
      const double tot = cnt==0 ? oProb : cnt;
      trans_sum += tot;
      trans->setProb(tot);
      trans->reset();
    }
    if( trans_sum > 0 ) {
      miter = outGoing.begin();
      for(; miter != mstop; ++miter) {
	RateTrans* trans = *miter;
	const double nProb = trans->getProb() / trans_sum;
	trans->setProb(nProb);
      }
    }
  }
}
