#include "TamuraNei.h"
#include "base.h"
#include <cmath>
#include <iostream>
#include "cidx.h"

using std::string;
using std::cout;
using std::endl;
using bioinfo::idx2c;
using bioinfo::c2idx;

TamuraNei::TamuraNei(const MarkovChain& mc, double tnTvRatio, double mRate) : 
  _tnTvRatio(tnTvRatio), _alphay(0), _alphar(0), _beta(0), _counts(4) {

    for(unsigned i = 0; i < 4; ++i) {
      const char ch = bioinfo::idx2c(i);
      const string str(1,ch);
      _counts[i] = mc.getVal(str,1);
    }
    init();
}

void
TamuraNei::print(std::ostream& os) const {
  for(unsigned i = 0; i < 4; ++i) {
    //const char ch = bioinfo::idx2c(i);
    os<<_counts[i]<<" ";
  }
  os<<_tnTvRatio<<endl;
}

void 
TamuraNei::init() {
  double pyFreq = 0, prFreq = 0;
  for(unsigned i = 0; i < 4; ++i) {
    const char c = idx2c(i);
    if( isPy(c) ) {
      pyFreq += getBaseFreq(c);
    } else if ( isPr(c) ) {
      prFreq += getBaseFreq(c);
    } else {
      assert(0);
    }
  }
  double freqr = prFreq;
  double freqy = pyFreq;
  const double freqa = getBaseFreq('a');
  const double freqc = getBaseFreq('c');
  const double freqg = getBaseFreq('g');
  const double freqt = getBaseFreq('t');

  double cR = _tnTvRatio;
  if( (freqr*freqy*cR)<((freqa*freqg)+(freqc*freqt))) {
    cR = ((freqa*freqg)-(freqc*freqt))/(freqr*freqy);
    cR += 0.1;
    if( (2*freqr*freqy*(1+cR)) < 1 ) {
      cR = 1/(2*freqr*freqy)+5;
    }
  }
  _beta = 1/(2*freqr*freqy*(1+cR));
  const double cRoll = freqy > 0 ? freqr/freqy : 100; // makes this HKY Model
  _alphay =  ((freqr*freqy*cR)-(freqa*freqg)-(freqc*freqt))/(2*(1+cR)*(freqy*freqa*freqg*cRoll+freqr*freqc*freqt));
  _alphar = cRoll*_alphay;
  if(freqr==0) {
    _alphay = 1/(2*freqc*freqt);
    _alphar = 0;
    _beta = 0;
  }
  if(freqy==0) {
    _alphay = 0;
    _alphar = 1/(2*freqa*freqg);
    _beta = 0;
  }
  assert(_alphay>=0&&_alphar>=0&&_beta>=0);
}


double
TamuraNei::getBaseFreq(char ch) const {
  const unsigned idx = bioinfo::c2idx(ch);
  return _counts[idx];
  //string str(1,ch);
  //return _mc.getVal(str,1);
}

#if 0
static int
kroneckerDelta(int i, int j) {
  return (i==j) ? 1 : 0;
}

double
TamuraNei::getProb(char tc, char fc, double blen) const {
  const unsigned ti = c2idx(tc);
  const unsigned fi = c2idx(fc);
  if( ti == 4 || fi == 4 ) {
    assert(0);
  }
  //blen *= _mRate;

  const int kd = kroneckerDelta(ti,fi);
  const double alpha = isPr(fc) ? _alphar : _alphay;

  // No Events
  const double nEvent = exp((-1*(alpha+_beta))*blen)*kd;
  // Transitions
  double sum = 0;

  const char base = idx2c(0);
  int lsb=sameBaseType(fc,base);
  sum += getBaseFreq(base)*lsb;

  const char base = idx2c(1);
  int lsb=sameBaseType(fc,base);
  sum += getBaseFreq(base)*lsb;

  const char base = idx2c(2);
  int lsb=sameBaseType(fc,base);
  sum += getBaseFreq(base)*lsb;

  const char base = idx2c(3);
  int lsb=sameBaseType(fc,base);
  sum += getBaseFreq(base)*lsb;

  const double toFreq = getBaseFreq(tc);
  const int sb=sameBaseType(tc,fc);
  const double pcnt = sum > 0 ? (toFreq/sum)*sb : 0;
  const double transVal = exp(-1*_beta*blen)*(1-exp(-1*alpha*blen))*pcnt;
  // Transversion (type 2)
  const double type2 = (1-exp(-1*_beta*blen))*toFreq;
  double res = nEvent + transVal + type2;
  if( res < 0 || res > 1 ) {
    assert(0);
  }
  assert(res>=0 && res<=1);
  return res;
}
#endif
