#include "TestModel.h"
#include "ParseHeader.h"
#include "ModelManager.h"
#include <algorithm>
#include <iostream>
#include <string>
#include <sstream>
#include <utility>

using std::pair;
using std::istringstream;
using std::vector;
using std::string;
using std::list;
using std::cout;
using std::endl;
typedef string::size_type size_type;

static double
getSp(double, double);
static double
getSn(double, double);

typedef pair<const SeqLabelVal*,const SeqLabelVal*> slvpair_t;

slvpair_t
findOverlap(const list<SeqLabelVal>::const_iterator& begin, const list<SeqLabelVal>::const_iterator& end, unsigned nuci ); 

static bool
isExonNuc(const SeqLabel& leftLabel, const SeqLabel& rghtLabel, signed pos); 

TestModel::TestModel(const string& id) : 
  _fp(ModelType::eCnameEnd,0), _tp(ModelType::eCnameEnd,0), _fn(ModelType::eCnameEnd,0), _tn(ModelType::eCnameEnd,0),
  _fp_splice(SeqLabel::eUndef,0), _tp_splice(SeqLabel::eUndef,0), _fn_splice(SeqLabel::eUndef,0), _tn_splice(SeqLabel::eUndef,0),
  _id(id),
  _fp_class(SeqLabel::eSpliceClassSize,0), _tp_class(SeqLabel::eSpliceClassSize,0), _fn_class(SeqLabel::eSpliceClassSize,0),
  _fp_nuc(0), _tp_nuc(0), _fn_nuc(0), _tn_nuc(0)
{
}

TestModel::TestModel(const TestModel& tm) :
  _fp(ModelType::eCnameEnd,0), _tp(ModelType::eCnameEnd,0), _fn(ModelType::eCnameEnd,0), _tn(ModelType::eCnameEnd,0),
  _fp_splice(SeqLabel::eUndef,0), _tp_splice(SeqLabel::eUndef,0), _fn_splice(SeqLabel::eUndef,0), _tn_splice(SeqLabel::eUndef,0),
  _id(tm._id),
  _fp_class(SeqLabel::eSpliceClassSize,0), _tp_class(SeqLabel::eSpliceClassSize,0), _fn_class(SeqLabel::eSpliceClassSize,0),
    _fp_nuc(0), _tp_nuc(0), _fn_nuc(0), _tn_nuc(0) {
  for(unsigned i = 0; i < _fp.size(); ++i) {
    _fp[i]=tm._fp[i];
    _tp[i]=tm._tp[i];
    _fn[i]=tm._fn[i];
    _tn[i]=tm._tn[i];
  }
  for(unsigned i = 0; i < _fp_class.size(); ++i) {
    _fp_splice[i]=tm._fp_splice[i];
    _tp_splice[i]=tm._tp_splice[i];
    _fn_splice[i]=tm._fn_splice[i];
    _tn_splice[i]=tm._tn_splice[i];
  }
  for(unsigned i = 0; i < _fp_class.size(); ++i) {
    _fp_class[i]=tm._fp_class[i];
    _tp_class[i]=tm._tp_class[i];
    _fn_class[i]=tm._fn_class[i];
  }
  _fp_nuc = tm._fp_nuc;
  _tp_nuc = tm._tp_nuc;
  _fn_nuc = tm._fn_nuc;
  _tn_nuc = tm._tn_nuc;
}


TestModel&
TestModel::operator+=(const TestModel& tm) {
  for(unsigned i = 0; i < _fp.size(); ++i) {
    _fp[i]+=tm._fp[i];
    _tp[i]+=tm._tp[i];
    _fn[i]+=tm._fn[i];
    _tn[i]+=tm._tn[i];
  }
  for(unsigned i = 0; i < _fp_class.size(); ++i) {
    _fp_class[i]+=tm._fp_class[i];
    _tp_class[i]+=tm._tp_class[i];
    _fn_class[i]+=tm._fn_class[i];
  }
  _fp_nuc += tm._fp_nuc;
  _tp_nuc += tm._tp_nuc;
  _fn_nuc += tm._fn_nuc;
  _tn_nuc += tm._tn_nuc;
  return *this;
}

std::ostream&
operator<<(std::ostream& os, const TestModel& tm) {
  double fp_don = 0, tp_don = 0, fn_don = 0, tn_don = 0;
  double fp_acc = 0, tp_acc = 0, fn_acc = 0, tn_acc = 0;
  double fp_prc = 0, tp_prc = 0, fn_prc = 0, tn_prc = 0;
  double fp_npc = 0, tp_npc = 0, fn_npc = 0, tn_npc = 0;
  double fp_int = 0, tp_int = 0, fn_int = 0, tn_int = 0;
  for(unsigned i = 0; i < ModelType::eCnameEnd; ++i) {
    if( ModelType::isDon((ModelType::cname_t)i) ) {
      fp_don += tm._fp[i];
      tp_don += tm._tp[i];
      fn_don += tm._fn[i];
      tn_don += tm._tn[i];
    } 
    else if( ModelType::isAcc((ModelType::cname_t)i) ) {
      fp_acc += tm._fp[i];
      tp_acc += tm._tp[i];
      fn_acc += tm._fn[i];
      tn_acc += tm._tn[i];
    } else if( ModelType::isPrcExon((ModelType::cname_t)i)) {
      fp_prc += tm._fp[i];
      tp_prc += tm._tp[i];
      fn_prc += tm._fn[i];
      tn_prc += tm._tn[i];
    } else if( ModelType::isNpcExon((ModelType::cname_t)i)) {
      fp_npc += tm._fp[i];
      tp_npc += tm._tp[i];
      fn_npc += tm._fn[i];
      tn_npc += tm._tn[i];
    } else if( ModelType::isIntron((ModelType::cname_t)i)) {
      fp_int += tm._fp[i];
      tp_int += tm._tp[i];
      fn_int += tm._fn[i];
      tn_int += tm._tn[i];
    }
  }
  const double sn_don = getSn(tp_don,fn_don);
  const double sp_don = getSp(tp_don,fp_don);
  const double sn_acc = getSn(tp_acc,fn_acc);
  const double sp_acc = getSp(tp_acc,fp_acc);
  const double sn_nuc = getSn(tm._tp_nuc,tm._fn_nuc);
  const double sp_nuc = getSp(tm._tp_nuc,tm._fp_nuc);

  const double sn_prc = getSn(tp_prc,fn_prc);
  const double sp_prc = getSp(tp_prc,fp_prc);
  const double sn_npc = getSn(tp_npc,fn_npc);
  const double sp_npc = getSp(tp_npc,fp_npc);
  const double sn_int = getSn(tp_int,fn_int);
  const double sp_int = getSp(tp_int,fp_int);

  //os<<"sn_cls: "<<sn_cls<<" sp_cls: "<<sp_cls<<" sn_nuc: "<<sn_nuc<<" sp_nuc: "<<sp_nuc<<endl;
  const double avg = (sn_don+sp_don+sn_acc+sp_acc+sn_nuc+sp_nuc)/6.0;
  os<<"sn_don: "<<sn_don<<" sp_don: "<<sp_don<<" sn_acc: "<<sn_acc<<" sp_acc: "<<sp_acc<<" sn_nuc: "<<sn_nuc<<" sp_nuc: "<<sp_nuc<<" avg "<<avg<<endl;
  os<<"sn_prc: "<<sn_prc<<" sp_prc: "<<sp_prc<<" sn_npc: "<<sn_npc<<" sp_npc: "<<sp_npc<<" sn_int: "<<sn_int<<" sp_int: "<<sp_int<<endl;
  for(unsigned i = 0; i < ModelType::eCnameEnd; ++i) {
    ModelType::cname_t cn = (ModelType::cname_t)i;
    os<<ModelType::cname2str(cn)<<" sn: "<<tm.getSn(i)<<" sp: "<<tm.getSp(i)<<endl;
  }
  return os;
}

double
TestModel::calcAvg() const {
  double fp_don = 0, tp_don = 0, fn_don = 0, tn_don = 0;
  double fp_acc = 0, tp_acc = 0, fn_acc = 0, tn_acc = 0;
  for(unsigned i = 0; i < ModelType::eCnameEnd; ++i) {
    if( ModelType::isDon((ModelType::cname_t)i) ) {
      fp_don += _fp[i];
      tp_don += _tp[i];
      fn_don += _fn[i];
      tn_don += _tn[i];
    } 
    else if( ModelType::isAcc((ModelType::cname_t)i) ) {
      fp_acc += _fp[i];
      tp_acc += _tp[i];
      fn_acc += _fn[i];
      tn_acc += _tn[i];
    }
  }
  double fp_cls=0, tp_cls = 0, fn_cls = 0;
  for(unsigned i = 0; i < _fp_class.size(); ++i ) {
    fp_cls += _fp_class[i];
    tp_cls += _tp_class[i];
    fn_cls += _fn_class[i];
  }
  const double sn_don = ::getSn(tp_don,fn_don);
  const double sp_don = ::getSp(tp_don,fp_don);
  const double sn_acc = ::getSn(tp_acc,fn_acc);
  const double sp_acc = ::getSp(tp_acc,fp_acc);
#if 0
  const double sn_cls = ::getSn(tp_cls,fn_cls);
  const double sp_cls = ::getSp(tp_cls,fp_cls);
#endif
  const double sn_nuc = ::getSn(_tp_nuc,_fn_nuc);
  const double sp_nuc = ::getSp(_tp_nuc,_fp_nuc);
  //os<<"sn_cls: "<<sn_cls<<" sp_cls: "<<sp_cls<<" sn_nuc: "<<sn_nuc<<" sp_nuc: "<<sp_nuc<<endl;
  const double avg = (sn_don+sp_don+sn_acc+sp_acc+sn_nuc+sp_nuc)/6.0;
  return avg;
}

double
TestModel::calcFscore() const {
  vector<double> best_sn(6,0), best_sp(6,0);
  best_sn[0] = getSn((int)ModelType::eAexonPrc);
  best_sp[0] = getSp((int)ModelType::eAexonPrc);
  best_sn[1] = getSn((int)ModelType::eCexonPrc);
  best_sp[1] = getSp((int)ModelType::eCexonPrc);

  best_sn[2] = getSn((int)ModelType::eAdon);
  best_sp[2] = getSp((int)ModelType::eAdon);
  best_sn[3] = getSn((int)ModelType::eCdon);
  best_sp[3] = getSp((int)ModelType::eAdon);

  best_sn[4] = getSn((int)ModelType::eAacc);
  best_sp[4] = getSp((int)ModelType::eAacc);
  best_sn[5] = getSn((int)ModelType::eCacc);
  best_sp[5] = getSp((int)ModelType::eCacc);
  double best_sum = 0, best_prd = 1.0;
  for(unsigned i = 0; i < best_sn.size(); ++i) {
    best_sum += (best_sn[i]+best_sp[i]);
    best_prd *= (best_sn[i]*best_sp[i]);
  }
  const double best_score = best_sum > 0 ? (2.0*(double)best_sn.size()*best_prd)/best_sum : 0;
  return best_score;
}

double
TestModel::getSp(int st) const {
  assert(st < (signed)_tp.size());
  const double sum = (_tp[st]+_fp[st]);
  const double denom = sum==0?1:sum;
  return _tp[st]/denom;
}

double
getSp(double tp, double fp) {
  const double sum = (tp+fp);
  const double denom = sum==0?1:sum;
  return tp/denom;
}


double
getSn(double tp, double fn) {
  const double sum = (tp+fn);
  const double denom = sum==0?1:sum;
  return tp/denom;
}

double
TestModel::getSn(int st) const {
  assert(st < (signed)_tp.size());
  const double sum = (_tp[st]+_fn[st]);
  const double denom = sum==0?1:sum;
  return _tp[st]/denom;
}

static bool
qlook(const SeqLabel& a, const SeqLabel& b) {
    bool aVal = SeqLabel::isDon(a.getLabel());
    bool bVal = SeqLabel::isDon(b.getLabel());
    return ( a.getEvent()==b.getEvent() && aVal==bVal);
}

#if 0
int
TestModel::label2idx(const SeqLabel& sl) {
  int idx = -1;
  if(SeqLabel::isDon(sl.getLabel()) ) {
    idx = sl.getEvent()==SeqLabel::eAlt ? 1 : 0;
  } else {
    idx = sl.getEvent()==SeqLabel::eAlt ? 3 : 2;
  }
  return idx;
}

int
TestModel::cname2idx(const ModelType::cname_t st) {
  int idx = -1;
  if( st == ModelType::eAacc ) {
    idx = 3;
  } else if( st == ModelType::eCacc) {
    idx = 2;
  } else if( st == ModelType::eAdon) {
    idx = 1;
  } else if( st == ModelType::eCdon) {
    idx = 0;
  } else {
    assert(0);
  }
  return idx;
}
#endif

list<SeqLabel>::const_iterator
my_find(list<SeqLabel>::const_iterator beg, list<SeqLabel>::const_iterator end, const SeqLabel& cl) {
  for(list<SeqLabel>::const_iterator iter = beg; iter != end; ++iter) {
    if( qlook(*iter,cl) ) {
      return iter;
    }
  }
  return end;
}

void
TestModel::score(const fasta_seq& fs, const list<SeqLabel>& pred, const list<SeqLabel>& ans) {
  list<SeqLabel>::const_iterator iter = pred.begin();
  list<SeqLabel>::const_iterator stop = pred.end();
  list<SeqLabel> tlst;
  for(; iter != stop; ++iter) {
    const SeqLabel& sl = *iter;
    if( sl.getLabel()==SeqLabel::eBeg || sl.getLabel()==SeqLabel::eEnd ) {
      continue;
    }
    list<SeqLabel>::const_iterator match = my_find(ans.begin(),ans.end(),sl);
    //const SeqLabel& am = *match;
    const int idx = (int)ModelType::sl2cnamePos(sl);
    if( match != ans.end()) {
      // true splice site
      assert(idx < (signed)_tp.size());
      ++_tp[idx];
    } else if( match == ans.end()) {
	// false splice site
      assert(idx < (signed)_fp.size());
      ++_fp[idx];
    }
    tlst.push_back(sl);
  }
  
  const AnnotationInfo* ai = AnnotationInfo::getInstance();
  ParseHeader ph(ai);
  ph.parse(fs);
  ExonDescrip& ed = ph.getExonDescrip();
  list<SeqLabel>::const_iterator ti = ans.begin(), ts = ans.end();
  const SeqLabel* prev = NULL;
  int aCnt = 0, dCnt = 0;
  for(; ti != ts; ++ti) {
    const SeqLabel& an = *ti;
    if( an.getLabel()==SeqLabel::eBeg ) { 
      prev = &an;
      continue;
    }
    if( SeqLabel::isDon(an.getLabel()) ) {
      ++dCnt;
    }
    if( SeqLabel::isAcc(an.getLabel()) ) {
      ++aCnt;
    }
    if( find(pred.begin(),pred.end(),an) == pred.end() ) {
      const SeqLabel& sl = *ti;
      const int idx = (int)ModelType::sl2cnamePos(sl);
      ++_fn[idx];
    }
    assert(prev);
  }
  ipair_t ip = ed.getStartStop();
  int beg = ans.front().getLeft()+1, end = ans.back().getRght();
  for(unsigned nuci = beg; nuci < (unsigned)end; nuci++) {
    const bool isPrc = ((signed)nuci >= ip.first && (signed)nuci <=ip.second);
    const tp_t pn = ModelManager::getCname(pred,nuci,isPrc);
    const tp_t tn = ModelManager::getCname(ans,nuci,isPrc);
    const bool isPredExon = ModelType::isPrcExon(pn.first);
    const bool isTrueExon = ModelType::isPrcExon(tn.first);
    const int tidx = (int)tn.first;
    const int pidx = (int)pn.first;
    assert(tidx < (signed)_tp.size());
    assert(pidx < (signed)_tp.size());
    if( pn.first == tn.first ) {
      // don't double count splice sites
      if( !ModelType::isSplice(pn.first) || (ModelType::isSplice(pn.first) && pn.second==0) ) {
	++_tp[tidx];
      }
    } else {
      if( !ModelType::isSplice(tn.first) || (ModelType::isSplice(tn.first) && tn.second==0) ) {
	++_fn[tidx];
      }
      if( !ModelType::isSplice(pn.first) || (ModelType::isSplice(pn.first) && pn.second==0) ) {
	++_fp[pidx];
      }
    }
    if( isTrueExon && isPredExon ) {
      ++_tp_nuc;
    } else if(isTrueExon && !isPredExon) {
      ++_fn_nuc;
    } else if(isPredExon && !isTrueExon) {
      ++_fp_nuc;
    } else {
      ++_tn_nuc;
    }
  }
#if 0
  sclass_t sc1 = SeqLabel::getSpliceClass(tlst);
  sclass_t sc2 = SeqLabel::eSpliceClassSize;

  const size_type tp = _id.find_first_of(':');
  const size_type tp2 = _id.find_first_of(':',tp+1);
  const size_type ll = _id.find('[');
  string cstr = _id.substr(tp2+2,ll-tp2-3);
  if( cstr == "MS" || cstr == "PT:MS" ) {
    if( aCnt > 1 ) {
      sc2 = SeqLabel::eMultipleAcceptor;
    } else {
      assert(dCnt>1);
      sc2 = SeqLabel::eMultipleDonor;
    }
  } else if( cstr == "IR"  || cstr == "PT:IR" ) {
    sc2 = SeqLabel::eIntronRetention;
  } else if( cstr == "SE" || cstr == "SE:MS" ) {
    sc2 = SeqLabel::eCassette;
  } else if( cstr == "CS" ) {
    sc2 = SeqLabel::eConstitutive;
  } else {
    assert(0);
  }
  if( sc1 == sc2 ) {
    ++_tp_class[sc1];
  } else {
    ++_fp_class[sc1];
    ++_fn_class[sc2];
  }
#endif
}

slvpair_t
findOverlap(const list<SeqLabelVal>::const_iterator& begin, const list<SeqLabelVal>::const_iterator & end, unsigned nuci) {
  const SeqLabelVal* left = NULL, *rght = NULL;
  list<SeqLabelVal>::const_iterator iter = begin;
  const SeqLabelVal* prev = &(*iter);
  ++iter;
  for(; iter != end; ++iter) {
    const SeqLabelVal& curr = *iter;
    if( isExonNuc(*prev,curr,nuci) ) {
      left = prev;
      rght = &curr;
      break;
    }
    prev = &curr;
  }
  return slvpair_t(left,rght);
}

bool
isExonNuc(const SeqLabel& leftLabel, const SeqLabel& rghtLabel, signed pos) {
  bool isExon = false;
  if( pos >= leftLabel.getLeft() && pos <= rghtLabel.getRght() ) {
  //if( !rghtLabel.isIn(pos) ) {
    // exon
    if( ( SeqLabel::isAcc(leftLabel.getLabel()) || leftLabel.getLabel() == SeqLabel::eBeg) && 
	( SeqLabel::isDon(rghtLabel.getLabel()) || rghtLabel.getLabel() == SeqLabel::eEnd) ) {
      isExon = true;
    }
    // retained intron intron
    else if( SeqLabel::isDon(leftLabel.getLabel()) &&
	     SeqLabel::isAcc(rghtLabel.getLabel()) ) {
      isExon = true;
    }
    // multiple donor region
    else if( SeqLabel::isDon(leftLabel.getLabel()) &&
	     SeqLabel::isDon(rghtLabel.getLabel()) ) {
      isExon = true;
    }
    // multiple acceptor region
    else if( SeqLabel::isAcc(leftLabel.getLabel()) &&
	     SeqLabel::isAcc(rghtLabel.getLabel()) ) {
      isExon = true;
    }
  }
  return isExon;
}
