#include "utils.h"
#include "ParamFile.h"
#include "OrgIdx.h"
#include "MarkovChain.h"
#include "AnnotationInfo.h"
#include "ParseHeader.h"
#include "stype.h"
#include "Histogram.h"
#include "base.h"
#include "cidx.h"

using bioinfo::idx2c;
using bioinfo::c2idx;
using std::ofstream;
using std::endl;
using std::cout;
using std::cerr;
using std::string;
using std::vector;
using std::list;

aahash_t GLOBAL_AA_HT;
char
codon2aa(const string& str) {
  return GLOBAL_AA_HT[str];
}


void
initGlobalAAHashTable() {
  GLOBAL_AA_HT["aaa"] = 'K';
  GLOBAL_AA_HT["aac"] = 'N';
  GLOBAL_AA_HT["aag"] = 'N';
  GLOBAL_AA_HT["aat"] = 'N';
  GLOBAL_AA_HT["aca"] = 'T';
  GLOBAL_AA_HT["acc"] = 'T';
  GLOBAL_AA_HT["acg"] = 'T';
  GLOBAL_AA_HT["act"] = 'T';
  GLOBAL_AA_HT["aga"] = 'R';
  GLOBAL_AA_HT["agc"] = 'S';
  GLOBAL_AA_HT["agg"] = 'R';
  GLOBAL_AA_HT["agt"] = 'T';
  GLOBAL_AA_HT["ata"] = 'I';
  GLOBAL_AA_HT["atc"] = 'I';
  GLOBAL_AA_HT["atg"] = 'M';
  GLOBAL_AA_HT["att"] = 'I';
  GLOBAL_AA_HT["caa"] = 'Q';
  GLOBAL_AA_HT["cac"] = 'H';
  GLOBAL_AA_HT["cag"] = 'Q';
  GLOBAL_AA_HT["cat"] = 'H';
  GLOBAL_AA_HT["cca"] = 'P';
  GLOBAL_AA_HT["ccc"] = 'P';
  GLOBAL_AA_HT["ccg"] = 'P';
  GLOBAL_AA_HT["cct"] = 'P';
  GLOBAL_AA_HT["cga"] = 'R';
  GLOBAL_AA_HT["cgc"] = 'R';
  GLOBAL_AA_HT["cgg"] = 'R';
  GLOBAL_AA_HT["cgt"] = 'R';
  GLOBAL_AA_HT["cta"] = 'L';
  GLOBAL_AA_HT["ctc"] = 'L';
  GLOBAL_AA_HT["ctg"] = 'L';
  GLOBAL_AA_HT["ctt"] = 'L';
  GLOBAL_AA_HT["gaa"] = 'E';
  GLOBAL_AA_HT["gac"] = 'D';
  GLOBAL_AA_HT["gag"] = 'E';
  GLOBAL_AA_HT["gat"] = 'D';
  GLOBAL_AA_HT["gca"] = 'A';
  GLOBAL_AA_HT["gcc"] = 'A';
  GLOBAL_AA_HT["gcg"] = 'A';
  GLOBAL_AA_HT["gct"] = 'A';
  GLOBAL_AA_HT["gga"] = 'G';
  GLOBAL_AA_HT["ggc"] = 'G';
  GLOBAL_AA_HT["ggg"] = 'G';
  GLOBAL_AA_HT["ggt"] = 'G';
  GLOBAL_AA_HT["gta"] = 'V';
  GLOBAL_AA_HT["gtc"] = 'V';
  GLOBAL_AA_HT["gtg"] = 'V';
  GLOBAL_AA_HT["gtt"] = 'V';
  GLOBAL_AA_HT["taa"] = 'Z';
  GLOBAL_AA_HT["tac"] = 'Y';
  GLOBAL_AA_HT["tag"] = 'Z';
  GLOBAL_AA_HT["tat"] = 'Y';
  GLOBAL_AA_HT["tca"] = 'S';
  GLOBAL_AA_HT["tcc"] = 'S';
  GLOBAL_AA_HT["tcg"] = 'S';
  GLOBAL_AA_HT["tct"] = 'S';
  GLOBAL_AA_HT["tga"] = 'Z';
  GLOBAL_AA_HT["tgc"] = 'C';
  GLOBAL_AA_HT["tgg"] = 'W';
  GLOBAL_AA_HT["tgt"] = 'C';
  GLOBAL_AA_HT["tta"] = 'L';
  GLOBAL_AA_HT["ttc"] = 'F';
  GLOBAL_AA_HT["ttg"] = 'L';
  GLOBAL_AA_HT["ttt"] = 'F';
}

void rghtFrm2LeftFrm(std::vector<int>& ftrack,int left,int rpos) {
  for(unsigned rframe = 0; rframe < 3; ++rframe) {
    int leftOver = 0;
    if( rframe == 2 ) {
      leftOver = 0;
    } else if( rframe == 1) {
      leftOver = 2;
    } else if( rframe == 0) {
      leftOver = 1;
    }
    const int diff = ((rpos-leftOver)-left) + 1;
    assert(diff > 0);
    ftrack[rframe] = (diff % 3);
  }
}

void
makelower(string& str) {
  for(unsigned i = 0; i < str.size(); ++i) {
    str[i] = tolower(str[i]);
  }
}

void
printAlignment(std::ostream& os, std::list<SeqLabel>::const_iterator iter, const std::list<SeqLabel>::const_iterator& stop, 
	       const std::vector<std::string*>& seqs) {
  ParamFile* pf = ParamFile::getInstance();
  const OrgIdx oidx = pf->getOrgIdx();
  std::list<SeqLabel>::const_iterator last = iter;
  ++iter;
  for(; iter != stop; ++iter) {
    const SeqLabel& left = *last;
    const SeqLabel& rght = *iter;
    os<<"From "<<left<<" to "<<rght<<endl;
    signed cnt = left.getRght()+1;
    const signed len = 80;
    while(cnt <= rght.getRght() && cnt < (signed)seqs[0]->size()) {
      int beg = cnt-1;
      int end = cnt+len>rght.getRght() ? rght.getRght() : cnt+len;
      end = end >= (signed)seqs[0]->size() ? (signed)seqs[0]->size()-1 : end;
      for(unsigned i = 0; i < seqs.size(); ++i) {
	if( seqs[i] ) {
	  assert(beg>=0);
	  int diff = end-beg;
	  assert((unsigned)beg < seqs[i]->size());
	  assert((unsigned)end < seqs[i]->size());
	  string sub = seqs[i]->substr(beg,diff);
	  string org = oidx.idx2str(i);
	  os<<org<<" "<<sub<<endl;
	}
      }
      os<<endl;
      cnt += len;
    }
    last = iter;
  }
}

fasta::const_iterator 
qfind(fasta::const_iterator iter, fasta::const_iterator stop, const string& hd) {
  for(; iter != stop; ++iter) {
    const fasta_seq& fs = **iter;
    if(fs.get_header()==hd)
      return iter;
  }
  return stop;
}

vector<const string*>
getMatchSeq(const fasta_seq& fs, const vector<fasta>& aln, 
	    const vector<string>& oid, const OrgIdx& oidx) {
  const string& hdr = fs.get_header();
  vector<const string*> svec(aln.size(),NULL);
  for(unsigned i = 0; i < aln.size(); ++i) {
    const fasta& fas = aln[i];
    const fasta::const_iterator fnd = qfind(fas.begin(),fas.end(),hdr);
    if( fnd != fas.end()) {
      const string id = oid[i];
      const int idx = oidx.fname2idx(id);
      const fasta_seq& tf = **fnd;
      svec[idx] = &(tf.get_seq());
    }
  }
  return svec;
}

typedef std::pair<unsigned,char> tpair_t;

class CmpOp {
public:
  bool operator()(const tpair_t& a, const tpair_t& b) const {
    return a.first < b.first;
  }
};

std::pair<double,int>
fillColumn(string& column,const vector<string*>& mal, unsigned i) {
  assert(i>=1);
  for(unsigned j = 0; j < mal.size(); ++j) {
    if(mal[j]) {
      const string& str = *mal[j];
      assert(i-1<str.size());
      column[j] = str[i-1];
      //assert(column[j] != '-');
    }
  }
  return std::pair<double,int>(0,0);
}

static void
partition_data_help(vector<fasta>& train, const fasta& masterlist) {
  const int num_p = train.size();
  //cout<<i<<" partition size: "<<partition<<endl;
  const int pwidth = masterlist.size()/num_p;
  for(int pari = 0; pari < num_p; ++pari) {
    assert((unsigned)pari < train.size());
    //fasta ltrain,ltest;
    fasta::const_iterator iter = masterlist.begin();
    const fasta::const_iterator stop = masterlist.end();
    int part = pari*pwidth, cnt = 0;
    //cout<<"pari: "<<pari<<" ps: "<<partsize[i]<<" "<<part<<" pwidth: "<<pwidth<<endl;
    for(; iter != stop; ++iter ) {
      if( cnt >= part && cnt <= (part+pwidth) ) {
	train[pari].push_back(*iter);
      }
      ++cnt;
    }
  }
}

void
partition_data(vector<fasta>& train, const fasta& masterlist1, const fasta& masterlist2) {
  partition_data_help(train,masterlist1);
  partition_data_help(train,masterlist2);
}


static void
partition_data_help1(vector<fasta>& train, vector<fasta>& test, const fasta& masterlist) {
  const int num_p = train.size();
  const int pwidth = (masterlist.size()/num_p) - 1;
  int remainder = masterlist.size() % num_p;
  //cout<<"data size: "<<num_p<<endl;
  int part = 0;
  for(int pari = 0; pari < num_p; ++pari) {
    assert((unsigned)pari < train.size() && (unsigned)pari < test.size());
    fasta::const_iterator iter = masterlist.begin();
    const fasta::const_iterator stop = masterlist.end();
    int cnt = 0;
    int twidth = pwidth;
    if( remainder > 0 ) {
      ++twidth;
      --remainder;
    }
    //cout<<"hmm: "<<pari<<" "<<pwidth<<" "<<part<<" "<<part+twidth<<endl;
    for(; iter != stop; ++iter, ++cnt ) {
      if( cnt < part || cnt > (part+twidth) ) {
	train[pari].push_back(*iter);
      } else {
	//cout<<"pari: "<<pari<<" range "<<cnt<<" to "<<part+pwidth<<" part "<<part<<" pwidth: "<<pwidth<<endl;
	test[pari].push_back(*iter);
      }
    }
    part += (twidth+1);
  }
}

void
partition_data(vector<fasta>& train, vector<fasta>& test, const fasta& masterlist1, const fasta& masterlist2) {
  partition_data_help1(train,test,masterlist1);
  partition_data_help1(train,test,masterlist2);
}

void
partition_train(const fasta& train, fasta& test1, fasta& test2, double pcnt) {
  const unsigned cLIM = 801;
  fasta::const_iterator iter = train.begin();
  const fasta::const_iterator stop = train.end();
  int altCnt = 0, conCnt = 0;
  for(; iter != stop; ++iter) {
    const fasta_seq& targ_fs = **iter;
    size_t pos = targ_fs.get_header().find("_CS_");
    const unsigned size = targ_fs.get_seq().size();
    if( pos != string::npos && size < cLIM ) {
      ++conCnt;
    } else if( size < cLIM ) {
      ++altCnt;
    }
  }
  double conLimit = pcnt*(double)conCnt;
  double altLimit = pcnt*(double)altCnt;
  cout<<"Total conTrain: "<<conLimit<<" total altLimit: "<<altLimit<<endl;
  iter = train.begin();
  for(; iter != stop; ++iter) {
    const fasta_seq& targ_fs = **iter;
    size_t pos = targ_fs.get_header().find("_CS_");
    const unsigned size = targ_fs.get_seq().size();
    if( pos != string::npos && size < cLIM ) {
      if( conLimit > 0 ) {
	test2.push_back(*iter);
	--conLimit;
      } else {
	test1.push_back(*iter);
      }
    } else if(size < cLIM) {
      if( altLimit > 0 ) {
	test2.push_back(*iter);
	--altLimit;
      } else {
	test1.push_back(*iter);
      }
    }
  }
}


static void
partition_data_help2(fasta& train, const fasta& masterlist, double pcnt) {
  const unsigned num = (int)(pcnt*(double)masterlist.size());
  fasta::const_iterator iter = masterlist.begin();
  const fasta::const_iterator stop = masterlist.end();
  unsigned cnt = 0;
  for(; iter != stop && cnt < num; ++iter, ++cnt ) {
    train.push_back(*iter);
  }
}

void
partition_data(fasta& train, const fasta& masterlist1, const fasta& masterlist2, double pcnt) {
  partition_data_help2(train,masterlist1,pcnt);
  partition_data_help2(train,masterlist2,pcnt);
}

const vector<string*>*
getMatchSeqPtr(const fasta_seq& targ) {
  ParamFile* pf = ParamFile::getInstance();
  vector<fasta>& alnSeq = pf->getAlignedSeq();
  const string& hdr = targ.get_header();
  vector<string*>* svecPtr = new vector<string*>(alnSeq.size(),NULL);
  vector<string*>& svec = *svecPtr;
  
  unsigned num_org = 0;
  for(unsigned i = 0; i < alnSeq.size(); ++i) {
    fasta& fas = alnSeq[i];
    fasta::iterator fnd = fas.find_header(hdr);
    if( fnd != fas.end()) {
      fasta_seq& tf = **fnd;
      svec[i] = &(tf.get_seq());
      ++num_org;
    }
  }
  return svecPtr;
}

const vector<string*>
getMatchSeq(const fasta_seq& targ) {
  ParamFile* pf = ParamFile::getInstance();
  vector<fasta>& alnSeq = pf->getAlignedSeq();
  const string& hdr = targ.get_header();
  vector<string*> svec(alnSeq.size(),NULL);
  unsigned num_org = 0;
  for(unsigned i = 0; i < alnSeq.size(); ++i) {
    fasta& fas = alnSeq[i];
    fasta::iterator fnd = fas.find_header(hdr);
    if( fnd != fas.end()) {
      fasta_seq& tf = **fnd;
      svec[i] = &(tf.get_seq());
      ++num_org;
    }
  }
#if 0
  for(unsigned i = 1; i <= svec[0]->size(); ++i) {
    assert(i>=1);
    vector<tpair_t> base_cnt(4);
    for(unsigned k = 0; k < base_cnt.size(); ++k) {
      base_cnt[k].first= 0;
      base_cnt[k].second = idx2c(k);
    }
    int avail=0,inDelCnt=0;
    bool fill_blanks=false;
    for(unsigned j = 0; j < svec.size(); ++j) {
      if(svec[j]) {
        const string& str = *svec[j];
        assert(i-1<str.size());
        const char ch = str[i-1];
        if(ch != '-' ) {
          ++base_cnt[ c2idx(ch) ].first;
        } else {
          fill_blanks=true;
          ++inDelCnt;
        }
        ++avail;
      }
    }
    if( fill_blanks ) {
      const string& tg = *svec[0];
      for(unsigned j = 0; j < svec.size(); ++j) {
        if(svec[j]) {
          string& str = *svec[j];
          const char ch = str[i-1];
          if( ch == '-' ) {
            std::sort(base_cnt.begin(),base_cnt.end(),CmpOp());
            const char most_common = base_cnt[3].second;
            const char least_common = base_cnt[0].second;
            const char second_least_common = base_cnt[1].second;
            char worse_base = least_common;
            if( base_cnt[0].first == base_cnt[1].first ) {
              worse_base = sameBaseType(most_common,least_common) ? second_least_common : least_common;
	    }
	    // prevent arbitrary insertion of stop codons
	    if( i>=3 ) {
	      if( isStop(tg[i-3],tg[i-2],worse_base) && !isStop(tg[i-3],tg[i-2],tg[i-1]) ) {
		for(unsigned tmp = 0; tmp < base_cnt.size(); ++tmp) {
		  char nbase = base_cnt[tmp].second;
		  if( !isStop(tg[i-3],tg[i-2],nbase) ) {
		    worse_base = nbase;
		    break;
		  }
		}
	      }
	    }
	    assert(worse_base != '-');
            str[i-1] = tolower(worse_base);
	    const char nch = str[i-1];
            for(unsigned z = 0; z < 4; ++z) {
              if(base_cnt[z].second == nch) {
                ++base_cnt[z].first;
              }
            }
          }
        }
      }
    }
  }
#endif
  return svec;
}


bool
isCons(unsigned i, const std::vector<string*>& svec, int) {
  //over write for now
  const int prev = 0;
  int start = i-prev;
  unsigned num_org = svec.size();
  bool isConsCol = true;
  if( start < 1 ) {
    isConsCol=false;
  } else {
  for(int iter = start; iter <= (signed)i; ++iter) {
    string column(num_org,'-');
    fillColumn(column,svec,iter /*i*/);
    for(unsigned k = 1; k < column.size(); ++k) {
      if( column[k] != column[0] && svec[k] ) {
	isConsCol=false;
	break;
      }
    }
    if( !isConsCol ) {
      break;
    }
  }
  }
  return isConsCol;
}

void
getPhases(vector<int>& ftrack, int lpos, int start) {
  assert(ftrack.size()==3);
  for(unsigned fidx = 0; fidx < 3; ++fidx) {
    const int sub = ((3-fidx)==3?0:3-fidx);
    const int startCodon = start >= 1 ? start : 1;
    if( lpos > (startCodon+sub) ) {
      const int diff = (lpos - (startCodon+sub)) % 3;
      if(diff == 0) 
	ftrack[fidx] = 0;
      else if(diff == 1)
	ftrack[fidx] = 1;
      else if(diff == 2)
	ftrack[fidx] = 2;
    } else if( lpos <= (startCodon+sub) ) {
      ftrack[fidx]=fidx;
    }
    assert(ftrack[fidx] >=0 && ftrack[fidx] <= 2);
  }
}

void
getPhases(vector<int>& ftrack, const SeqLabel& tl, int start) {
  const int startCodon = start > 1 ? start : 1;
  for(unsigned fidx = 0; fidx < 3; ++fidx) {
    const int sub = ((3-fidx)==3?0:3-fidx);
    ftrack[fidx] = fidx;
    if( tl.getExonBnd() > (startCodon+sub) ) {
      if( SeqLabel::isDon(tl.getLabel()) && start < tl.getLeft() ) {
	const int pos = tl.getExonBnd()+7;
	const int diff = (pos - (startCodon+sub)) % 3;
	ftrack[fidx] = diff;
      } else if( SeqLabel::isAcc(tl.getLabel()) && start < tl.getLeft() ) {
	const int pos = tl.getExonBnd()+2;
	const int diff = (pos-(startCodon+sub)) % 3;
	ftrack[fidx] = diff;
      }
    }
    assert(ftrack[fidx] >=0 && ftrack[fidx] <= 2);
  }
}

double
calcPcntId(const string& str1, const string& str2, int left, int rght) {
  if( rght >= (signed)str1.size())
    rght = (signed)str1.size()-1;
  assert( left>=1 && rght < (signed)str1.size() && rght < (signed) str2.size());
  unsigned id=0;
  for(int i = left; i <= rght; ++i) {
    if( str1[i-1]==str2[i-1] ) {
      ++id;
    }
  }
  int len = (rght-left) + 1;
  return (double)id/(double)len;
}

void
transfer(fasta& alt, fasta& con) {
   list<fasta::iterator> rem;
   fasta::iterator iter = alt.begin();
   const fasta::iterator stop = alt.end();
   for(; iter != stop; ++iter) {
      const fasta_seq& fs = **iter;
      fasta_seq* fsPtr = *iter;
      const string& head =fs.get_header();
      if( strstr(head.c_str(),"_CS_") ) {
         con.push_back(fsPtr);
         rem.push_back(iter);
      }
   }
   list<fasta::iterator>::iterator i1=rem.begin();
   const list<fasta::iterator>::iterator s1=rem.end();
   for(; i1 != s1; ++i1) {
      alt.erase(*i1);
   }
}

bool
hasGcDon(const string& tseq, const list<SeqLabel>& lst) {
  const list<SeqLabel>::const_iterator end = lst.end();
  list<SeqLabel>::const_iterator iter = lst.begin();
  bool isGc=false;
  for(; iter != end; iter++) {
    const SeqLabel* cslv = &(*iter);
    if( SeqLabel::isDon(cslv->getLabel()) ) {
      int pos = cslv->getExonBnd()+1;
      if( tolower(tseq[pos-1])=='g' && tolower(tseq[pos]) == 'c' ) {
	isGc=true;
	break;
      }
    }
  }
  return isGc;
}

bool
hasNonConsensusSplice(const string& tseq, const list<SeqLabel>& lst) {
  const list<SeqLabel>::const_iterator end = lst.end();
  list<SeqLabel>::const_iterator iter = lst.begin();
  bool isNon=false;
  for(; iter != end; iter++) {
    const SeqLabel* cslv = &(*iter);
    if( SeqLabel::isDon(cslv->getLabel()) ) {
      int pos = cslv->getExonBnd()+1;
      if( tolower(tseq[pos-1])!='g' || tolower(tseq[pos])!= 't' ) {
	isNon=true;
      }
    }
    if( SeqLabel::isAcc(cslv->getLabel()) ) {
      int pos = cslv->getExonBnd()-1;
      if( tolower(tseq[pos-2])!='a' || tolower(tseq[pos-1])!= 'g' ) {
	isNon=true;
      }
    }
  }
  return isNon;
}

int
frame2exalt(int frame, int pos, int startc) {
  vector<int> opt(3,0);
  getPhases(opt,pos,startc);
  int res=-1;
  for(unsigned match = 0; match < opt.size(); ++match) {
    if(opt[match]==frame) {
      res=(int)match;
      break;
    }
  }
  return res;
}


bool
isPrcCheck(int pos, int start, int stop) {
  return (pos >= (start+3) && pos <= stop);
}


void usage(const char* prog) {
  cout<<endl;
  cout<<"------------------------------------------"<<endl;
  cout<<"ExAlt: predict alternatively spliced exons"<<endl;
  cout<<"------------------------------------------"<<endl<<endl;
  cout<<endl<<"usage: "<<prog;
  cout<<" -p \"data file\" -m \"model file\" -o \"gff output file\" [-i header] [-fevhs]\n";
  cout<<"\t-p\tdata input file lists filenames of the target multi fasta file and informant species\n";
  cout<<"\t-m\tprediction model file\n";
  cout<<"\t-o\toutput file name\n";
  cout<<"\t-f\tuse gene structure frame information\n";
  cout<<"\t-e\tuse input exon in prediction\n";
  cout<<"\t-s\tpredict at most one exon per sequence\n";
  cout<<"\t-l\tprint debugging information to log file\n";
  cout<<"\t-i\theader - run exalt only on fasta record with input header value\n";
  cout<<"\t-h\tprint usage message\n";
  cout<<"\t-v\tversion information\n";
  cout<<"\n\tA typical execution of exalt from the command line:\n";
  cout<<"\t"<<prog<<" -p paramFile.txt -m drosophila.models -o predictions.gff -e \n\n";
}
