/**
 * @file bio_bit_traits.h
**/ 
#ifndef _BIO_BIT_TRAITS_H
#define _BIO_BIT_TRAITS_H

#include "bit_char_traits.h"


namespace bioinfo 
{

  /**
   * @class bio_bit_traits
   * @brief "char_traits" specification for bit encoded bio sequence characters
   * @ingroup bio_string
   * 
   * The type of characters expected are either nucleic acid codes or 
   * amino acid codes. Code sizes are expected to vary depending on the application.
   * The initial case defaults to the case where only 4 character codes are used:
   * <b>a</b>denine,<b>c</b>ytosine,<b>g</b>uanine,<b>t</b>hymine.
   * template parameter NB defines the number of bits per character.
   *
   */
  template<size_t NB>
  class bio_bit_traits : public std::char_traits<std::bitset<NB> > {
  public:
    typedef typename std::char_traits<std::bitset<NB> >::char_type char_type; /*!< single character type */
    typedef typename std::char_traits<char>::char_type basic_char_type; /*!< reference to <i>char</i> type */

    /**
     * @return number of bits used to encode a single nucleic acid character
     */
    static std::size_t 
    bit_size() { 
      return 2;
    }

    /**
     * @param char_type = single character of sequence
     * @return single character code defined by right most bit_size() bits in char_type
     * @see bit_size()
     */
    inline static basic_char_type char_code(char_type);

    /**
     * @param basic_char_type single character (ascii code)
     * @return custom bit encoding of single character
     */
    inline static std::size_t bit_code(basic_char_type);

    /**
     * @param basic_char_type = single character
     * @return true if basic_char_type is recognized
     */
    inline static bool valid_code(basic_char_type);
  };


  /**
   *  na_nmer_traits defines character traits for a 
   *  bit compressed n-length nucleic acid character sequence
   *  where a series of these character sequences are overlapping
   * @ingroup bio_string
   * 
   * Extends bio_bit_traits
   *  As an example, if we are working with the sequence 
   *  "atcataatc", where n = 3, the string is defined as:
   *  atc tca cat ata taa aat atc (trailing tc will be ignored)
   *  for n=3, NB=na_min_traits::bit_size() * n.
   *  For this example NB=8 and memeroy storage
   *  is the same as the original sequence.
   * @see bio_bit_traits, nmer_string
   */
  template<size_t NB>
  class na_nmer_overlap : public bio_bit_traits<NB> {
  };
}


/**
 * namespace std operators
 * */
namespace std
{
  /** @{ */
  /** default operators **/
  template <class _CharT, class _Traits, typename _NaTraits, size_t _Nb>
  basic_ostream<_CharT, _Traits>&
    operator<<(basic_ostream<_CharT, _Traits>&, const basic_string<bitset<_Nb>, _NaTraits>&);

  template<class _CharT, class _Traits, class _NaTraits, size_t _Nb>
  basic_istream<_CharT, _Traits>&
  operator>>(basic_istream<_CharT, _Traits>&, basic_string<bitset<_Nb>, _NaTraits>&); 

  /** template specializations **/
  template<class _CharT, class _Traits, size_t __Nb>
  basic_istream<_CharT, _Traits>&
  operator>>(basic_istream<_CharT, _Traits>&, basic_string<bitset<__Nb>, bioinfo::na_nmer_overlap<__Nb> >&);

  template <class _CharT, class _Traits, size_t _Nb>
  basic_ostream<_CharT, _Traits>&
  operator<<(basic_ostream<_CharT, _Traits>&, const basic_string<bitset<_Nb>, bioinfo::na_nmer_overlap<_Nb> >&);

  template<size_t __Nb>
  bool
  operator==(const basic_string<bitset<__Nb>, bioinfo::na_nmer_overlap<__Nb> >&, const string&);
  /** @} */
}


#include "impl/bio_bit_traits.hpp"
#include "impl/na_nmer_overlap_stream.hpp"

#endif // _BIO_BIT_TRAITS_H
